diff -ruNpa u-boot-2010.12/board/ti/beagle/config.mk u-boot-2010.12.new/board/ti/beagle/config.mk
--- u-boot-2010.12/board/ti/beagle/config.mk	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/board/ti/beagle/config.mk	2011-03-01 11:48:06.811549998 +1100
@@ -31,3 +31,7 @@
 
 # For use with external or internal boots.
 CONFIG_SYS_TEXT_BASE = 0x80008000
+
+CONFIG_USB_DFU_VENDOR=0x1457
+CONFIG_USB_DFU_PRODUCT=0x5119
+CONFIG_USB_DFU_REVISION=0x0230
diff -ruNpa u-boot-2010.12/drivers/serial/usbtty.c u-boot-2010.12.new/drivers/serial/usbtty.c
--- u-boot-2010.12/drivers/serial/usbtty.c	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/drivers/serial/usbtty.c	2011-03-01 11:48:06.811549998 +1100
@@ -22,6 +22,8 @@
  */
 
 #include <common.h>
+#include <usb_dfu_descriptors.h>
+#include <usb_dfu.h>
 #include <config.h>
 #include <circbuf.h>
 #include <stdio_dev.h>
@@ -101,7 +103,7 @@ extern struct usb_string_descriptor **us
 static unsigned short rx_endpoint = 0;
 static unsigned short tx_endpoint = 0;
 static unsigned short interface_count = 0;
-static struct usb_string_descriptor *usbtty_string_table[STR_COUNT];
+static struct usb_string_descriptor *usbtty_string_table[NUM_STRINGS];
 
 /* USB Descriptor Strings */
 static u8 wstrLang[4] = {4,USB_DT_STRING,0x9,0x4};
@@ -151,6 +153,10 @@ struct acm_config_desc {
 	/* Slave Interface */
 	struct usb_interface_descriptor data_class_interface;
 	struct usb_endpoint_descriptor data_endpoints[NUM_ENDPOINTS-1];
+#ifdef CONFIG_USBD_DFU
+	struct usb_interface_descriptor uif_dfu;
+	struct usb_dfu_func_descriptor func_dfu;
+#endif
 } __attribute__((packed));
 
 static struct acm_config_desc acm_configuration_descriptors[NUM_CONFIGS] = {
@@ -161,7 +167,11 @@ static struct acm_config_desc acm_config
 			.bDescriptorType = USB_DT_CONFIG,
 			.wTotalLength =
 				cpu_to_le16(sizeof(struct acm_config_desc)),
+#ifdef CONFIG_USBD_DFU
+			.bNumInterfaces = NUM_ACM_INTERFACES + 1,
+#else
 			.bNumInterfaces = NUM_ACM_INTERFACES,
+#endif
 			.bConfigurationValue = 1,
 			.iConfiguration = STR_CONFIG,
 			.bmAttributes =
@@ -260,6 +270,11 @@ static struct acm_config_desc acm_config
 				.bInterval		= 0xFF,
 			},
 		},
+#ifdef CONFIG_USBD_DFU
+		/* Interface 3 */
+		.uif_dfu = DFU_RT_IF_DESC,
+		.func_dfu = DFU_FUNC_DESC,
+#endif
 	},
 };
 
@@ -370,7 +385,7 @@ static int fill_buffer (circbuf_t * buf)
 void usbtty_poll (void);
 
 /* utility function for converting char* to wide string used by USB */
-static void str2wide (char *str, u16 * wide)
+void str2wide (char *str, u16 * wide)
 {
 	int i;
 	for (i = 0; i < strlen (str) && str[i]; i++){
@@ -639,6 +654,9 @@ static void usbtty_init_instances (void)
 	device_instance->bus = bus_instance;
 	device_instance->configurations = NUM_CONFIGS;
 	device_instance->configuration_instance_array = config_instance;
+#ifdef CONFIG_USBD_DFU
+	dfu_init_instance(device_instance);
+#endif
 
 	/* initialize bus instance */
 	memset (bus_instance, 0, sizeof (struct usb_bus_instance));
diff -ruNpa u-boot-2010.12/drivers/serial/usbtty.h u-boot-2010.12.new/drivers/serial/usbtty.h
--- u-boot-2010.12/drivers/serial/usbtty.h	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/drivers/serial/usbtty.h	2011-03-01 11:48:06.811549998 +1100
@@ -84,4 +84,10 @@
 #define STR_CTRL_INTERFACE	0x06
 #define STR_COUNT		0x07
 
+#ifdef CONFIG_USBD_DFU
+#define NUM_STRINGS		DFU_STR_COUNT
+#else
+#define NUM_STRINGS		STR_COUNT
+#endif
+
 #endif
diff -ruNpa u-boot-2010.12/drivers/usb/gadget/core.c u-boot-2010.12.new/drivers/usb/gadget/core.c
--- u-boot-2010.12/drivers/usb/gadget/core.c	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/drivers/usb/gadget/core.c	2011-03-01 11:48:06.811549998 +1100
@@ -31,6 +31,7 @@
 
 #include <malloc.h>
 #include <usbdevice.h>
+#include <usb_dfu.h>
 
 #define MAX_INTERFACES 2
 
@@ -209,6 +210,10 @@ struct usb_alternate_instance *usbd_devi
  */
 struct usb_device_descriptor *usbd_device_device_descriptor (struct usb_device_instance *device, int port)
 {
+#ifdef CONFIG_USBD_DFU
+	if (device->dfu_state != DFU_STATE_appIDLE)
+		return device->dfu_dev_desc;
+#endif
 	return (device->device_descriptor);
 }
 
@@ -229,6 +234,10 @@ struct usb_configuration_descriptor *usb
 	if (!(configuration_instance = usbd_device_configuration_instance (device, port, configuration))) {
 		return NULL;
 	}
+#ifdef CONFIG_USBD_DFU
+	if (device->dfu_state != DFU_STATE_appIDLE)
+		return (&device->dfu_cfg_desc->ucfg);
+#endif
 	return (configuration_instance->configuration_descriptor);
 }
 
@@ -250,6 +259,13 @@ struct usb_interface_descriptor *usbd_de
 	if (!(interface_instance = usbd_device_interface_instance (device, port, configuration, interface))) {
 		return NULL;
 	}
+#ifdef CONFIG_USBD_DFU
+	if (device->dfu_state != DFU_STATE_appIDLE) {
+		if (alternate < 0 || alternate >= DFU_NUM_ALTERNATES)
+			return NULL;
+		return &device->dfu_cfg_desc->uif[alternate];
+	}
+#endif
 	if ((alternate < 0) || (alternate >= interface_instance->alternates)) {
 		return NULL;
 	}
@@ -680,4 +696,7 @@ void usbd_device_event_irq (struct usb_d
 		/* usbdbg("calling device->event"); */
 		device->event(device, event, data);
 	}
+#ifdef CONFIG_USBD_DFU
+	dfu_event(device, event, data);
+#endif
 }
diff -ruNpa u-boot-2010.12/drivers/usb/gadget/ep0.c u-boot-2010.12.new/drivers/usb/gadget/ep0.c
--- u-boot-2010.12/drivers/usb/gadget/ep0.c	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/drivers/usb/gadget/ep0.c	2011-03-01 11:48:06.811549998 +1100
@@ -52,6 +52,11 @@
 
 #include <common.h>
 #include <usbdevice.h>
+DECLARE_GLOBAL_DATA_PTR;
+
+#ifdef CONFIG_USBD_DFU
+#include <usb_dfu.h>
+#endif
 
 #if 0
 #define dbg_ep0(lvl,fmt,args...) serial_printf("[%s] %s:%d: "fmt"\n",__FILE__,__FUNCTION__,__LINE__,##args)
@@ -198,7 +203,7 @@ static int ep0_get_descriptor (struct us
 	urb->actual_length = 0;
 	cp = (char*)urb->buffer;
 
-	dbg_ep0 (2, "%s", USBD_DEVICE_DESCRIPTORS (descriptor_type));
+	dbg_ep0 (2, "%s (%d)", USBD_DEVICE_DESCRIPTORS (descriptor_type), descriptor_type);
 
 	switch (descriptor_type) {
 	case USB_DESCRIPTOR_TYPE_DEVICE:
@@ -276,7 +281,28 @@ static int ep0_get_descriptor (struct us
 	case USB_DESCRIPTOR_TYPE_ENDPOINT:
 		serial_printf("USB_DESCRIPTOR_TYPE_ENDPOINT - error not implemented\n");
 		return -1;
+	/* This really means "Class Specific Descriptor #1 == USB_DT_DFU */
 	case USB_DESCRIPTOR_TYPE_HID:
+#ifdef CONFIG_USBD_DFU
+		{
+			int bNumInterface =
+				le16_to_cpu(urb->device_request.wIndex);
+
+			/* In runtime mode, we only respond to the DFU INTERFACE,
+			 * whereas in DFU mode, we respond for all intrfaces */
+			if (device->dfu_state != DFU_STATE_appIDLE &&
+			    device->dfu_state != DFU_STATE_appDETACH ||
+			    bNumInterface == CONFIG_USBD_DFU_INTERFACE) {
+
+				/* copy descriptor for this device */
+				copy_config (urb, &device->dfu_cfg_desc->func_dfu,
+				     sizeof (struct usb_dfu_func_descriptor),
+				     max);
+			} else {
+				return -1;
+            }
+		}
+#else /* CONFIG_USBD_DFU */
 		{
 			serial_printf("USB_DESCRIPTOR_TYPE_HID - error not implemented\n");
 			return -1;	/* unsupported at this time */
@@ -304,6 +330,7 @@ static int ep0_get_descriptor (struct us
 				     max);
 #endif
 		}
+#endif /* CONFIG_USBD_DFU */
 		break;
 	case USB_DESCRIPTOR_TYPE_REPORT:
 		{
@@ -405,6 +432,24 @@ int ep0_recv_setup (struct urb *urb)
 		 le16_to_cpu (request->wLength),
 		 USBD_DEVICE_REQUESTS (request->bRequest));
 
+#ifdef CONFIG_USBD_DFU
+	if ((request->bmRequestType & 0x3f) == USB_TYPE_DFU &&
+	     (device->dfu_state != DFU_STATE_appIDLE ||
+	      le16_to_cpu(request->wIndex) == CONFIG_USBD_DFU_INTERFACE)) {
+		int rc = dfu_ep0_handler(urb);
+		switch (rc) {
+		case DFU_EP0_NONE:
+		case DFU_EP0_UNHANDLED:
+			break;
+		case DFU_EP0_ZLP:
+		case DFU_EP0_DATA:
+			return 0;
+		case DFU_EP0_STALL:
+			return -1;
+		}
+	}
+#endif /* CONFIG_USBD_DFU */
+
 	/* handle USB Standard Request (c.f. USB Spec table 9-2) */
 	if ((request->bmRequestType & USB_REQ_TYPE_MASK) != 0) {
 		if(device->device_state <= STATE_CONFIGURED){
@@ -582,7 +627,8 @@ int ep0_recv_setup (struct urb *urb)
 			device->interface = le16_to_cpu (request->wIndex);
 			device->alternate = le16_to_cpu (request->wValue);
 			/*dbg_ep0(2, "set interface: %d alternate: %d", device->interface, device->alternate); */
-			serial_printf ("DEVICE_SET_INTERFACE.. event?\n");
+			usbd_device_event_irq(device, DEVICE_SET_INTERFACE,
+						(request->wIndex << 16 | request->wValue));
 			return 0;
 
 		case USB_REQ_GET_STATUS:
diff -ruNpa u-boot-2010.12/drivers/usb/gadget/Makefile u-boot-2010.12.new/drivers/usb/gadget/Makefile
--- u-boot-2010.12/drivers/usb/gadget/Makefile	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/drivers/usb/gadget/Makefile	2011-03-01 11:48:06.811549998 +1100
@@ -33,6 +33,7 @@ else
 ifdef CONFIG_USB_DEVICE
 COBJS-y += core.o
 COBJS-y += ep0.o
+COBJS-y += usbdfu.o
 COBJS-$(CONFIG_OMAP1510) += omap1510_udc.o
 COBJS-$(CONFIG_OMAP1610) += omap1510_udc.o
 COBJS-$(CONFIG_MPC885_FAMILY) += mpc8xx_udc.o
diff -ruNpa u-boot-2010.12/drivers/usb/gadget/usbdfu.c u-boot-2010.12.new/drivers/usb/gadget/usbdfu.c
--- u-boot-2010.12/drivers/usb/gadget/usbdfu.c	1970-01-01 10:00:00.000000000 +1000
+++ u-boot-2010.12.new/drivers/usb/gadget/usbdfu.c	2011-03-01 11:48:06.811549998 +1100
@@ -0,0 +1,1053 @@
+/*
+ * (C) 2007 by OpenMoko, Inc.
+ * Author: Harald Welte <laforge@openmoko.org>
+ *
+ * based on existing SAM7DFU code from OpenPCD:
+ * (C) Copyright 2006 by Harald Welte <hwelte@hmw-consulting.de>
+ *
+ * See file CREDITS for list of people who contributed to this
+ * project.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+ * MA 02111-1307 USA
+ *
+ * TODO:
+ * - make NAND support reasonably self-contained and put in apropriate
+ *   ifdefs
+ * - add some means of synchronization, i.e. block commandline access
+ *   while DFU transfer is in progress, and return to commandline once
+ *   we're finished
+ * - add VERIFY support after writing to flash
+ * - sanely free() resources allocated during first uppload/download
+ *   request when aborting
+ * - sanely free resources when another alternate interface is selected
+ *
+ * Maybe:
+ * - add something like uImage or some other header that provides CRC
+ *   checking?
+ * - make 'dnstate' attached to 'struct usb_device_instance'
+ */
+
+#include <config.h>
+#if defined(CONFIG_USBD_DFU)
+
+#include <common.h>
+DECLARE_GLOBAL_DATA_PTR;
+
+#include <malloc.h>
+#include <linux/types.h>
+#include <linux/list.h>
+#include <asm/errno.h>
+#include <usbdevice.h>
+#include <usb_dfu.h>
+#include <usb_dfu_descriptors.h>
+#include <usb_dfu_trailer.h>
+
+#include <nand.h>
+#include <jffs2/load_kernel.h>
+int mtdparts_init(void);
+extern struct list_head devices;
+
+#include "../../serial/usbtty.h"			/* for STR_* defs */
+
+#define RET_NOTHING	0
+#define RET_ZLP		1
+#define RET_STALL	2
+
+volatile enum dfu_state *system_dfu_state; /* for 3rd parties */
+
+
+struct dnload_state {
+	nand_info_t *nand;
+	struct part_info *part;
+	unsigned int part_net_size;	/* net sizee (excl. bad blocks) of part */
+
+	nand_erase_options_t erase_opts;
+	nand_write_options_t write_opts;
+	nand_read_options_t read_opts;
+
+	unsigned char *ptr;	/* pointer to next empty byte in buffer */
+	unsigned int off;	/* offset of current erase page in flash chip */
+	unsigned char *buf;	/* pointer to allocated erase page buffer */
+
+	/* unless doing an atomic transfer, we use the static buffer below.
+	 * This saves us from having to clean up dynamic allications in the
+	 * various error paths of the code.  Also, it will always work, no
+	 * matter what the memory situation is. */
+	unsigned char _buf[0x20000];	/* FIXME: depends flash page size */
+};
+
+static struct dnload_state _dnstate;
+
+static int dfu_trailer_matching(const struct uboot_dfu_trailer *trailer)
+{
+	if (trailer->magic != UBOOT_DFU_TRAILER_MAGIC ||
+	    trailer->version != UBOOT_DFU_TRAILER_V1 ||
+	    trailer->vendor != CONFIG_USBD_VENDORID ||
+	    (trailer->product != CONFIG_USBD_PRODUCTID_CDCACM &&
+	     trailer->product != CONFIG_USBD_PRODUCTID_GSERIAL))
+		return 0;
+#ifdef CONFIG_REVISION_TAG
+	if (trailer->revision != get_board_rev())
+		return 0;
+#endif
+
+	return 1;
+}
+
+static struct part_info *get_partition_nand(int idx)
+{
+#if 0
+	struct mtd_device *dev;
+	struct part_info *part;
+	struct list_head *pentry;
+	int i;
+
+	if (mtdparts_init())
+		return NULL;
+	if (list_empty(&devices))
+		return NULL;
+
+	dev = list_entry(devices.next, struct mtd_device, link);
+	i = 0;
+	list_for_each(pentry, &dev->parts) {
+		if (i == idx)  {
+			part = list_entry(pentry, struct part_info, link);
+			return part;
+		}
+		i++;
+	}
+
+#endif
+	return NULL;
+}
+
+#define LOAD_ADDR ((unsigned char *)0x81000000)
+
+#if 0
+static int initialize_ds_nand(struct usb_device_instance *dev, struct dnload_state *ds)
+{
+	ds->part = get_partition_nand(dev->alternate - 1);
+	if (!ds->part) {
+		printf("DFU: unable to find partition %u\b", dev->alternate-1);
+   			dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errADDRESS;
+		return RET_STALL;
+	}
+	ds->nand = &nand_info[ds->part->dev->id->num];
+	ds->off = ds->part->offset;
+	ds->part_net_size = nand_net_part_size(ds->part);
+
+	if (ds->nand->erasesize > sizeof(ds->_buf)) {
+		printf("*** Warning - NAND ERASESIZE bigger than static buffer\n");
+		ds->buf = malloc(ds->nand->erasesize);
+		if (!ds->buf) {
+			printf("DFU: can't allocate %u bytes\n", ds->nand->erasesize);
+   			dev->dfu_state = DFU_STATE_dfuERROR;
+			dev->dfu_status = DFU_STATUS_errADDRESS;
+			return RET_STALL;
+		}
+	} else
+		ds->buf = ds->_buf;
+
+	ds->ptr = ds->buf;
+
+	memset(&ds->read_opts, 0, sizeof(ds->read_opts));
+
+	memset(&ds->erase_opts, 0, sizeof(ds->erase_opts));
+	ds->erase_opts.quiet = 1;
+	/* FIXME: do this more dynamic */
+	if (!strcmp(ds->part->name, "rootfs"))
+		ds->erase_opts.jffs2 = 1;
+
+	memset(&ds->write_opts, 0, sizeof(ds->write_opts));
+	ds->write_opts.pad = 1;
+	ds->write_opts.blockalign = 1;
+	ds->write_opts.quiet = 1;
+
+	debug("initialize_ds_nand(dev=%p, ds=%p): ", dev, ds);
+	debug("nand=%p, ptr=%p, buf=%p, off=0x%x\n", ds->nand, ds->ptr, ds->buf, ds->off);
+
+	return RET_NOTHING;
+}
+
+static int erase_flash_verify_nand(struct urb *urb, struct dnload_state *ds,
+				   unsigned long erasesize, unsigned long size)
+{
+	struct usb_device_instance *dev = urb->device;
+	int rc;
+
+	debug("erase_flash_verify_nand(urb=%p, ds=%p, erase=0x%x size=0x%x)\n",
+		urb, ds, erasesize, size);
+
+	if (erasesize == ds->nand->erasesize) {
+		/* we're only writing a single block and need to
+		 * do bad block skipping / offset adjustments our own */
+		while (ds->nand->block_isbad(ds->nand, ds->off)) {
+			debug("SKIP_ONE_BLOCK(0x%08x)!!\n", ds->off);
+			ds->off += ds->nand->erasesize;
+		}
+	}
+
+	/* we have finished one eraseblock, flash it */
+	ds->erase_opts.offset = ds->off;
+	ds->erase_opts.length = erasesize;
+	debug("Erasing 0x%x bytes @ offset 0x%x (jffs=%u)\n",
+		ds->erase_opts.length, ds->erase_opts.offset,
+		ds->erase_opts.jffs2);
+	rc = nand_erase_opts(ds->nand, &ds->erase_opts);
+	if (rc) {
+		debug("Error erasing\n");
+	    	dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errERASE;
+		return RET_STALL;
+	}
+
+	ds->write_opts.buffer = ds->buf;
+	ds->write_opts.length = size;
+	ds->write_opts.offset = ds->off;
+	debug("Writing 0x%x bytes @ offset 0x%x\n", size, ds->off);
+	rc = nand_write_opts(ds->nand, &ds->write_opts);
+	if (rc) {
+		debug("Error writing\n");
+  		dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errWRITE;
+		return RET_STALL;
+	}
+
+	ds->off += size;
+	ds->ptr = ds->buf;
+
+	/* FIXME: implement verify! */
+	return RET_NOTHING;
+}
+
+static int erase_tail_clean_nand(struct urb *urb, struct dnload_state *ds)
+{
+	struct usb_device_instance *dev = urb->device;
+	int rc;
+
+	ds->erase_opts.offset = ds->off;
+	ds->erase_opts.length = ds->part->size - (ds->off - ds->part->offset);
+	debug("Erasing tail of 0x%x bytes @ offset 0x%x (jffs=%u)\n",
+		ds->erase_opts.length, ds->erase_opts.offset,
+		ds->erase_opts.jffs2);
+	rc = nand_erase_opts(ds->nand, &ds->erase_opts);
+	if (rc) {
+		printf("Error erasing tail\n");
+	    	dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errERASE;
+		return RET_STALL;
+	}
+
+	ds->off += ds->erase_opts.length; /* for consistency */
+
+	return RET_NOTHING;
+}
+
+/* Read the next erase blcok from NAND into buffer */
+static int read_next_nand(struct urb *urb, struct dnload_state *ds, int len)
+{
+	struct usb_device_instance *dev = urb->device;
+	int rc;
+
+	ds->read_opts.buffer = ds->buf;
+	ds->read_opts.length = len;
+	ds->read_opts.offset = ds->off;
+	ds->read_opts.quiet = 1;
+
+	debug("Reading 0x%x@0x%x to 0x%08p\n", len, ds->off, ds->buf);
+	rc = nand_read_opts(ds->nand, &ds->read_opts);
+	if (rc) {
+		debug("Error reading\n");
+  		dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errWRITE;
+		return RET_STALL;
+	}
+	ds->off += len;
+	ds->ptr = ds->buf;
+
+	return RET_NOTHING;
+}
+#endif
+
+static int handle_dnload(struct urb *urb, u_int16_t val, u_int16_t len, int first)
+{
+	struct usb_device_instance *dev = urb->device;
+	struct dnload_state *ds = &_dnstate;
+	unsigned int actual_len = len;
+	unsigned int remain_len;
+	unsigned long size;
+	int rc;
+
+	debug("download(len=%u, first=%u) ", len, first);
+
+	if (len > CONFIG_USBD_DFU_XFER_SIZE) {
+		/* Too big. Not that we'd really care, but it's a
+		 * DFU protocol violation */
+		debug("length exceeds flash page size ");
+	    	dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errADDRESS;
+		return RET_STALL;
+	}
+
+	if (first && dev->alternate != 0) {
+		/* Make sure that we have a valid mtd partition table */
+		char *mtdp = getenv("mtdparts");
+		if (!mtdp)
+			run_command("dynpart", 0);
+	}
+
+	if (len == 0) {
+		debug("zero-size write -> MANIFEST_SYNC ");
+		dev->dfu_state = DFU_STATE_dfuMANIFEST_SYNC;
+
+		/* cleanup */
+		switch (dev->alternate) {
+			char buf[12];
+		case 0:
+			sprintf(buf, "%lx", ds->ptr - ds->buf);
+			setenv("filesize", buf);
+			ds->ptr = ds->buf;
+			break;
+#if 0
+		case 1:
+			if (ds->ptr >
+			    ds->buf + sizeof(struct uboot_dfu_trailer)) {
+				struct uboot_dfu_trailer trailer;
+				dfu_trailer_mirror(&trailer, ds->ptr);
+				if (!dfu_trailer_matching(&trailer)) {
+					printf("DFU TRAILER NOT OK\n");
+					dev->dfu_state = DFU_STATE_dfuERROR;
+					dev->dfu_status = DFU_STATUS_errTARGET;
+					return RET_STALL;
+				}
+
+				rc = erase_flash_verify_nand(urb, ds,
+							ds->part->size,
+							ds->part_net_size);
+				/* re-write dynenv marker in OOB */
+				run_command("dynenv set u-boot_env", 0);
+			}
+			ds->nand = NULL;
+			free(ds->buf);
+			ds->ptr = ds->buf = ds->_buf;
+#endif
+			break;
+		default:
+#if 0
+			rc = 0;
+			if (ds->ptr > ds->buf)
+				rc = erase_flash_verify_nand(urb, ds,
+							ds->nand->erasesize,
+							ds->nand->erasesize);
+			/* rootfs partition */
+			if (!rc && !strcmp(ds->part->name, "rootfs"))
+				rc = erase_tail_clean_nand(urb, ds);
+
+			ds->nand = NULL;
+#endif
+			break;
+		}
+
+		return RET_ZLP;
+	}
+
+	if (urb->actual_length != len) {
+		debug("urb->actual_length(%u) != len(%u) ?!? ",
+			urb->actual_length, len);
+	    	dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errADDRESS;
+		return RET_STALL;
+	}
+
+	if (first && ds->buf && ds->buf != ds->_buf && ds->buf != LOAD_ADDR) {
+		free(ds->buf);
+		ds->buf = ds->_buf;
+	}
+
+	switch (dev->alternate) {
+	case 0:
+		if (first) {
+			printf("Starting DFU DOWNLOAD to RAM (0x%08p)\n",
+				LOAD_ADDR);
+			ds->buf = LOAD_ADDR;
+			ds->ptr = ds->buf;
+		}
+
+		memcpy(ds->ptr, urb->buffer, len);
+		ds->ptr += len;
+		break;
+#if 0
+	case 1:
+		if (first) {
+			rc = initialize_ds_nand(dev, ds);
+			if (rc)
+				return rc;
+			ds->buf = malloc(ds->part_net_size);
+			if (!ds->buf) {
+				printf("No memory for atomic buffer!!\n");
+				dev->dfu_state = DFU_STATE_dfuERROR;
+				dev->dfu_status = DFU_STATUS_errUNKNOWN;
+				return RET_STALL;
+			}
+			ds->ptr = ds->buf;
+			printf("Starting Atomic DFU DOWNLOAD to partition '%s'\n",
+				ds->part->name);
+		}
+
+		remain_len = (ds->buf + ds->part_net_size) - ds->ptr;
+		if (remain_len < len) {
+			len = remain_len;
+			printf("End of write exceeds partition end\n");
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			dev->dfu_status = DFU_STATUS_errADDRESS;
+			return RET_STALL;
+		}
+		memcpy(ds->ptr, urb->buffer, len);
+		ds->ptr += len;
+		break;
+#endif
+	default:
+#if 0
+		if (first) {
+			rc = initialize_ds_nand(dev, ds);
+			if (rc)
+				return rc;
+			printf("Starting DFU DOWNLOAD to partition '%s'\n",
+				ds->part->name);
+		}
+
+		size = ds->nand->erasesize;
+		remain_len = ds->buf + size - ds->ptr;
+		if (remain_len < len)
+			actual_len = remain_len;
+
+		memcpy(ds->ptr, urb->buffer, actual_len);
+		ds->ptr += actual_len;
+
+		/* check partition end */
+		if (ds->off + (ds->ptr - ds->buf) > ds->part->offset + ds->part->size) {
+			printf("End of write exceeds partition end\n");
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			dev->dfu_status = DFU_STATUS_errADDRESS;
+			return RET_STALL;
+		}
+
+		if (ds->ptr >= ds->buf + size) {
+			rc = erase_flash_verify_nand(urb, ds,
+						     ds->nand->erasesize,
+						     ds->nand->erasesize);
+			if (rc)
+				return rc;
+			/* copy remainder of data into buffer */
+			memcpy(ds->ptr, urb->buffer + actual_len, len - actual_len);
+			ds->ptr += (len - actual_len);
+		}
+#endif
+		break;
+	}
+
+	return RET_ZLP;
+}
+
+static int handle_upload(struct urb *urb, u_int16_t val, u_int16_t len, int first)
+{
+	struct usb_device_instance *dev = urb->device;
+	struct dnload_state *ds = &_dnstate;
+	unsigned int remain;
+	int rc;
+
+	debug("upload(val=0x%02x, len=%u, first=%u) ", val, len, first);
+
+	if (len > CONFIG_USBD_DFU_XFER_SIZE) {
+		/* Too big */
+		dev->dfu_state = DFU_STATE_dfuERROR;
+		dev->dfu_status = DFU_STATUS_errADDRESS;
+		//udc_ep0_send_stall();
+		debug("Error: Transfer size > CONFIG_USBD_DFU_XFER_SIZE ");
+		return -EINVAL;
+	}
+
+	switch (dev->alternate) {
+	case 0:
+		if (first) {
+			printf("Starting DFU Upload of RAM (0x%08p)\n",
+				LOAD_ADDR);
+			ds->ptr = ds->buf;
+		}
+
+		/* FIXME: end at some more dynamic point */
+		if (ds->ptr + len > LOAD_ADDR + 0x200000)
+			len = (LOAD_ADDR + 0x200000) - ds->ptr;
+
+		memcpy(urb->buffer, ds->ptr, len);
+		urb->actual_length = len;
+		ds->ptr += len;
+		break;
+	default:
+#if 1
+		printf("Rejecting NAND upload.\n");
+#else
+		if (first) {
+			rc = initialize_ds_nand(dev, ds);
+			if (rc)
+				return -EINVAL;
+			printf("Starting DFU Upload of partition '%s'\n",
+				ds->part->name);
+		}
+
+		if (len > ds->nand->erasesize) {
+			printf("We don't support transfers bigger than %u\n",
+				ds->nand->erasesize);
+			len = ds->nand->erasesize;
+		}
+
+		/* limit length to whatever number of bytes is left in
+		 * this partition */
+		remain = (ds->part->offset + ds->part->size) - ds->off;
+		if (len > remain)
+			len = remain;
+
+		rc = read_next_nand(urb, ds, len);
+		if (rc)
+			return -EINVAL;
+
+		debug("uploading %u bytes ", len);
+		urb->buffer = ds->buf;
+		urb->actual_length = len;
+		break;
+#endif
+	}
+
+	debug("returning len=%u\n", len);
+	return len;
+}
+
+static void handle_getstatus(struct urb *urb, int max)
+{
+	struct usb_device_instance *dev = urb->device;
+	struct dfu_status *dstat = (struct dfu_status *) urb->buffer;
+
+	debug("getstatus ");
+
+	if (!urb->buffer || urb->buffer_length < sizeof(*dstat)) {
+		debug("invalid urb! ");
+		return;
+	}
+
+	switch (dev->dfu_state) {
+	case DFU_STATE_dfuDNLOAD_SYNC:
+	case DFU_STATE_dfuDNBUSY:
+#if 0
+		if (fsr & AT91C_MC_PROGE) {
+			debug("errPROG ");
+			dev->dfu_status = DFU_STATUS_errPROG;
+			dev->dfu_state = DFU_STATE_dfuERROR;
+		} else if (fsr & AT91C_MC_LOCKE) {
+			debug("errWRITE ");
+			dev->dfu_status = DFU_STATUS_errWRITE;
+			dev->dfu_state = DFU_STATE_dfuERROR;
+		} else if (fsr & AT91C_MC_FRDY) {
+#endif
+			debug("DNLOAD_IDLE ");
+			dev->dfu_state = DFU_STATE_dfuDNLOAD_IDLE;
+#if 0
+		} else {
+			debug("DNBUSY ");
+			dev->dfu_state = DFU_STATE_dfuDNBUSY;
+		}
+#endif
+		break;
+	case DFU_STATE_dfuMANIFEST_SYNC:
+		break;
+	default:
+		//return;
+		break;
+	}
+
+	/* send status response */
+	dstat->bStatus = dev->dfu_status;
+	dstat->bState = dev->dfu_state;
+	dstat->iString = 0;
+	/* FIXME: set dstat->bwPollTimeout */
+	urb->actual_length = MIN(sizeof(*dstat), max);
+
+	/* we don't need to explicitly send data here, will
+	 * be done by the original caller! */
+}
+
+static void handle_getstate(struct urb *urb, int max)
+{
+	debug("getstate ");
+
+	if (!urb->buffer || urb->buffer_length < sizeof(u_int8_t)) {
+		debug("invalid urb! ");
+		return;
+	}
+
+	urb->buffer[0] = urb->device->dfu_state & 0xff;
+	urb->actual_length = sizeof(u_int8_t);
+}
+
+#ifndef CONFIG_USBD_PRODUCTID_DFU
+#define CONFIG_USBD_PRODUCTID_DFU CONFIG_USBD_PRODUCTID_CDCACM
+#endif
+
+static const struct usb_device_descriptor dfu_dev_descriptor = {
+	.bLength		= USB_DT_DEVICE_SIZE,
+	.bDescriptorType	= USB_DT_DEVICE,
+	.bcdUSB			= 0x0100,
+	.bDeviceClass		= 0x00,
+	.bDeviceSubClass	= 0x00,
+	.bDeviceProtocol	= 0x00,
+	.bMaxPacketSize0	= EP0_MAX_PACKET_SIZE,
+	.idVendor		= CONFIG_USBD_VENDORID,
+	.idProduct		= CONFIG_USBD_PRODUCTID_DFU,
+	.bcdDevice		= 0x0000,
+	.iManufacturer		= DFU_STR_MANUFACTURER,
+	.iProduct		= DFU_STR_PRODUCT,
+	.iSerialNumber		= DFU_STR_SERIAL,
+	.bNumConfigurations	= 0x01,
+};
+
+static struct _dfu_desc dfu_cfg_descriptor = {
+	.ucfg = {
+		.bLength		= USB_DT_CONFIG_SIZE,
+		.bDescriptorType	= USB_DT_CONFIG,
+		.wTotalLength		= USB_DT_CONFIG_SIZE +
+					  DFU_NUM_ALTERNATES * USB_DT_INTERFACE_SIZE +
+					  USB_DT_DFU_SIZE,
+		.bNumInterfaces		= 1,
+		.bConfigurationValue	= 1,
+		.iConfiguration		= DFU_STR_CONFIG,
+		.bmAttributes		= BMATTRIBUTE_RESERVED,
+		.bMaxPower		= 50,
+	},
+	.func_dfu = DFU_FUNC_DESC,
+};
+
+int dfu_ep0_handler(struct urb *urb)
+{
+	int rc, ret = RET_NOTHING;
+	u_int8_t req = urb->device_request.bRequest;
+	u_int16_t val = urb->device_request.wValue;
+	u_int16_t len = urb->device_request.wLength;
+	struct usb_device_instance *dev = urb->device;
+
+	debug("dfu_ep0(req=0x%x, val=0x%x, len=%u) old_state = %u ",
+		req, val, len, dev->dfu_state);
+
+	switch (dev->dfu_state) {
+	case DFU_STATE_appIDLE:
+		switch (req) {
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		case USB_REQ_DFU_DETACH:
+			dev->dfu_state = DFU_STATE_appDETACH;
+			ret = RET_ZLP;
+			goto out;
+			break;
+		default:
+			ret = RET_STALL;
+		}
+		break;
+	case DFU_STATE_appDETACH:
+		switch (req) {
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_appIDLE;
+			ret = RET_STALL;
+			goto out;
+			break;
+		}
+		/* FIXME: implement timer to return to appIDLE */
+		break;
+	case DFU_STATE_dfuIDLE:
+		switch (req) {
+		case USB_REQ_DFU_DNLOAD:
+            printf("Got DNLOAD req in dfuIDLE, len = %d\n", len);
+			if (len == 0) {
+				dev->dfu_state = DFU_STATE_dfuERROR;
+				ret = RET_STALL;
+				goto out;
+			}
+			dev->dfu_state = DFU_STATE_dfuDNLOAD_SYNC;
+			ret = handle_dnload(urb, val, len, 1);
+			break;
+		case USB_REQ_DFU_UPLOAD:
+			dev->dfu_state = DFU_STATE_dfuUPLOAD_IDLE;
+			handle_upload(urb, val, len, 1);
+			break;
+		case USB_REQ_DFU_ABORT:
+			/* no zlp? */
+			ret = RET_ZLP;
+			break;
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		case USB_REQ_DFU_DETACH:
+			/* Proprietary extension: 'detach' from idle mode and
+			 * get back to runtime mode in case of USB Reset.  As
+			 * much as I dislike this, we just can't use every USB
+			 * bus reset to switch back to runtime mode, since at
+			 * least the Linux USB stack likes to send a number of resets
+			 * in a row :( */
+			dev->dfu_state = DFU_STATE_dfuMANIFEST_WAIT_RST;
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			goto out;
+			break;
+		}
+		break;
+	case DFU_STATE_dfuDNLOAD_SYNC:
+		switch (req) {
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			/* FIXME: state transition depending on block completeness */
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			goto out;
+		}
+		break;
+	case DFU_STATE_dfuDNBUSY:
+		switch (req) {
+		case USB_REQ_DFU_GETSTATUS:
+			/* FIXME: only accept getstatus if bwPollTimeout
+			 * has elapsed */
+			handle_getstatus(urb, len);
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			goto out;
+		}
+		break;
+	case DFU_STATE_dfuDNLOAD_IDLE:
+		switch (req) {
+		case USB_REQ_DFU_DNLOAD:
+			dev->dfu_state = DFU_STATE_dfuDNLOAD_SYNC;
+			ret = handle_dnload(urb, val, len, 0);
+			break;
+		case USB_REQ_DFU_ABORT:
+			dev->dfu_state = DFU_STATE_dfuIDLE;
+			ret = RET_ZLP;
+			break;
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			break;
+		}
+		break;
+	case DFU_STATE_dfuMANIFEST_SYNC:
+		switch (req) {
+		case USB_REQ_DFU_GETSTATUS:
+			/* We're MainfestationTolerant */
+			dev->dfu_state = DFU_STATE_dfuIDLE;
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			break;
+		}
+		break;
+	case DFU_STATE_dfuMANIFEST:
+		/* we should never go here */
+		dev->dfu_state = DFU_STATE_dfuERROR;
+		ret = RET_STALL;
+		break;
+	case DFU_STATE_dfuMANIFEST_WAIT_RST:
+		/* we should never go here */
+		break;
+	case DFU_STATE_dfuUPLOAD_IDLE:
+		switch (req) {
+		case USB_REQ_DFU_UPLOAD:
+			/* state transition if less data then requested */
+			rc = handle_upload(urb, val, len, 0);
+			if (rc >= 0 && rc < len)
+				dev->dfu_state = DFU_STATE_dfuIDLE;
+			break;
+		case USB_REQ_DFU_ABORT:
+			dev->dfu_state = DFU_STATE_dfuIDLE;
+			/* no zlp? */
+			ret = RET_ZLP;
+			break;
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			break;
+		}
+		break;
+	case DFU_STATE_dfuERROR:
+		switch (req) {
+		case USB_REQ_DFU_GETSTATUS:
+			handle_getstatus(urb, len);
+			break;
+		case USB_REQ_DFU_GETSTATE:
+			handle_getstate(urb, len);
+			break;
+		case USB_REQ_DFU_CLRSTATUS:
+			dev->dfu_state = DFU_STATE_dfuIDLE;
+			dev->dfu_status = DFU_STATUS_OK;
+			/* no zlp? */
+			ret = RET_ZLP;
+			break;
+		default:
+			dev->dfu_state = DFU_STATE_dfuERROR;
+			ret = RET_STALL;
+			break;
+		}
+		break;
+	default:
+		return DFU_EP0_UNHANDLED;
+		break;
+	}
+
+out:
+	debug("new_state = %u, ret = %u\n", dev->dfu_state, ret);
+
+	switch (ret) {
+	case RET_ZLP:
+		//udc_ep0_send_zlp();
+		urb->actual_length = 0;
+		return DFU_EP0_ZLP;
+		break;
+	case RET_STALL:
+		//udc_ep0_send_stall();
+		return DFU_EP0_STALL;
+		break;
+	case RET_NOTHING:
+		break;
+	}
+
+	return DFU_EP0_DATA;
+}
+
+void str2wide (char *str, u16 * wide);
+static struct usb_string_descriptor *create_usbstring(char *string)
+{
+	struct usb_string_descriptor *strdesc;
+	int size = sizeof(*strdesc) + strlen(string)*2;
+
+	if (size > 255)
+		return NULL;
+
+	strdesc = malloc(size);
+	if (!strdesc)
+		return NULL;
+
+	strdesc->bLength = size;
+	strdesc->bDescriptorType = USB_DT_STRING;
+	str2wide(string, strdesc->wData);
+
+	return strdesc;
+}
+
+
+static void dfu_init_strings(struct usb_device_instance *dev)
+{
+	int i;
+	struct usb_string_descriptor *strdesc;
+
+	strdesc = create_usbstring(CONFIG_DFU_CFG_STR);
+	usb_strings[DFU_STR_CONFIG] = strdesc;
+
+	for (i = 0; i < DFU_NUM_ALTERNATES; i++) {
+		if (i == 0) {
+			strdesc = create_usbstring(CONFIG_DFU_ALT0_STR);
+		} else {
+			struct part_info *part = get_partition_nand(i-1);
+
+			if (part)
+				strdesc = create_usbstring(part->name);
+			else
+				strdesc =
+				    create_usbstring("undefined partition");
+		}
+		if (!strdesc)
+			continue;
+		usb_strings[STR_COUNT+i+1] = strdesc;
+	}
+}
+
+
+#ifdef CONFIG_NAND_DYNPART
+
+void dfu_update_strings(void)
+{
+	int i;
+
+	if (!system_dfu_state) {
+		printf("NASTY SURPRISE: system_dfu_state not set\n");
+		return;
+	}
+
+	for (i = 1; i != DFU_NUM_ALTERNATES; i++) {
+		struct part_info *part = get_partition_nand(i-1);
+		struct usb_string_descriptor *strdesc, **slot;
+
+		if (part)
+			strdesc = create_usbstring(part->name);
+		else
+			strdesc = create_usbstring("undefined partition");
+		if (!strdesc)
+			continue;
+		slot = usb_strings+STR_COUNT+i+1;
+		if (*slot)
+			free(*slot);
+		*slot = strdesc;
+	}
+}
+
+#endif /* CONFIG_NAND_DYNPART */
+
+
+int dfu_init_instance(struct usb_device_instance *dev)
+{
+	int i;
+
+	for (i = 0; i != DFU_NUM_ALTERNATES; i++) {
+		struct usb_interface_descriptor *uif =
+		    dfu_cfg_descriptor.uif+i;
+
+		uif->bLength		= USB_DT_INTERFACE_SIZE;
+		uif->bDescriptorType	= USB_DT_INTERFACE;
+		uif->bAlternateSetting	= i;
+		uif->bInterfaceClass	= 0xfe;
+		uif->bInterfaceSubClass	= 1;
+		uif->bInterfaceProtocol	= 2;
+		uif->iInterface		= DFU_STR_ALT(i);
+	}
+
+	dev->dfu_dev_desc = &dfu_dev_descriptor;
+	dev->dfu_cfg_desc = &dfu_cfg_descriptor;
+	dev->dfu_state = DFU_STATE_appIDLE;
+	dev->dfu_status = DFU_STATUS_OK;
+
+	if (system_dfu_state)
+		printf("SURPRISE: system_dfu_state is already set\n");
+	system_dfu_state = &dev->dfu_state;
+
+	dfu_init_strings(dev);
+
+	return 0;
+}
+
+static int stdout_switched;
+
+/* event handler for usb device state events */
+void dfu_event(struct usb_device_instance *device,
+	       usb_device_event_t event, int data)
+{
+	char *out;
+
+	switch (event) {
+	case DEVICE_RESET:
+		switch (device->dfu_state) {
+		case DFU_STATE_appDETACH:
+			device->dfu_state = DFU_STATE_dfuIDLE;
+			out = getenv("stdout");
+			if (out && !strcmp(out, "usbtty")) {
+				setenv("stdout", "vga");
+				setenv("stderr", "vga");
+				stdout_switched = 1;
+			}
+			printf("DFU: Switching to DFU Mode\n");
+			break;
+		case DFU_STATE_dfuMANIFEST_WAIT_RST:
+			device->dfu_state = DFU_STATE_appIDLE;
+			printf("DFU: Switching back to Runtime mode\n");
+			if (stdout_switched) {
+				setenv("stdout", "usbtty");
+				setenv("stderr", "usbtty");
+				stdout_switched = 0;
+			}
+			break;
+		default:
+			break;
+		}
+		break;
+	case DEVICE_CONFIGURED:
+	case DEVICE_DE_CONFIGURED:
+		debug("SET_CONFIGURATION(%u) ", device->configuration);
+		/* fallthrough */
+	case DEVICE_SET_INTERFACE:
+		debug("SET_INTERFACE(%u,%u) old_state = %u ",
+			device->interface, device->alternate,
+			device->dfu_state);
+		switch (device->dfu_state) {
+		case DFU_STATE_appIDLE:
+		case DFU_STATE_appDETACH:
+		case DFU_STATE_dfuIDLE:
+		case DFU_STATE_dfuMANIFEST_WAIT_RST:
+			/* do nothing, we're fine */
+			break;
+		case DFU_STATE_dfuDNLOAD_SYNC:
+		case DFU_STATE_dfuDNBUSY:
+		case DFU_STATE_dfuDNLOAD_IDLE:
+		case DFU_STATE_dfuMANIFEST:
+			device->dfu_state = DFU_STATE_dfuERROR;
+			device->dfu_status = DFU_STATUS_errNOTDONE;
+			/* FIXME: free malloc()ed buffer! */
+			break;
+		case DFU_STATE_dfuMANIFEST_SYNC:
+		case DFU_STATE_dfuUPLOAD_IDLE:
+		case DFU_STATE_dfuERROR:
+			device->dfu_state = DFU_STATE_dfuERROR;
+			device->dfu_status = DFU_STATUS_errUNKNOWN;
+			break;
+		}
+		debug("new_state = %u\n", device->dfu_state);
+		break;
+	default:
+		break;
+	}
+}
+#endif /* CONFIG_USBD_DFU */
diff -ruNpa u-boot-2010.12/drivers/usb/musb/musb_udc.c u-boot-2010.12.new/drivers/usb/musb/musb_udc.c
--- u-boot-2010.12/drivers/usb/musb/musb_udc.c	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/drivers/usb/musb/musb_udc.c	2011-03-01 11:28:36.501550000 +1100
@@ -1,925 +1,692 @@
 /*
- * Copyright (c) 2009 Wind River Systems, Inc.
- * Tom Rix <Tom.Rix@windriver.com>
- *
- * This file is a rewrite of the usb device part of
- * repository git.omapzoom.org/repo/u-boot.git, branch master,
- * file cpu/omap3/fastboot.c
- *
- * This is the unique part of its copyright :
- *
- * -------------------------------------------------------------------------
- *
- * (C) Copyright 2008 - 2009
- * Windriver, <www.windriver.com>
- * Tom Rix <Tom.Rix@windriver.com>
- *
- * -------------------------------------------------------------------------
- *
- * The details of connecting the device to the uboot usb device subsystem
- * came from the old omap3 repository www.sakoman.net/u-boot-omap3.git,
- * branch omap3-dev-usb, file drivers/usb/usbdcore_musb.c
- *
- * This is the unique part of its copyright :
- *
- * -------------------------------------------------------------------------
- *
- * (C) Copyright 2008 Texas Instruments Incorporated.
+ * (C) Copyright 2009 Texas Instruments Incorporated.
  *
  * Based on
  * u-boot OMAP1510 USB drivers (drivers/usbdcore_omap1510.c)
  * twl4030 init based on linux (drivers/i2c/chips/twl4030_usb.c)
  *
- * Author: Diego Dompe (diego.dompe@ridgerun.com)
- *         Atin Malaviya (atin.malaviya@gmail.com)
- *
- * -------------------------------------------------------------------------
- *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of
- * the License, or (at your option) any later version.
+ * Author:	Diego Dompe (diego.dompe@ridgerun.com)
+ *		Atin Malaviya (atin.malaviya@gmail.com)
+ *		Matthew Warton (mwarton@ok-labs.com)
+ *		Bernard Blackham (b-musb@largestprime.net)
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
- * MA 02111-1307 USA
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  */
 
 #include <common.h>
+
+#include <asm/io.h>
+#include <asm/arch/clocks.h>
+#include <asm/arch/clocks_omap3.h>
+#include <asm/arch/sys_proto.h>
 #include <usb/musb_udc.h>
 #include "../gadget/ep0.h"
-#include "musb_core.h"
 #if defined(CONFIG_USB_OMAP3)
 #include "omap3.h"
-#elif defined(CONFIG_USB_AM35X)
-#include "am35x.h"
 #elif defined(CONFIG_USB_DAVINCI)
 #include "davinci.h"
 #endif
 
-/* Define MUSB_DEBUG for debugging */
-/* #define MUSB_DEBUG */
-#include "musb_debug.h"
-
-#define MAX_ENDPOINT 15
-
-#define GET_ENDPOINT(dev,ep)						\
-(((struct usb_device_instance *)(dev))->bus->endpoint_array + ep)
-
-#define SET_EP0_STATE(s)						\
-do {									\
-	if ((0 <= (s)) && (SET_ADDRESS >= (s))) {			\
-		if ((s) != ep0_state) {					\
-			if ((debug_setup) && (debug_level > 1))		\
-				serial_printf("INFO : Changing state "  \
-					      "from %s to %s in %s at " \
-					      "line %d\n",		\
-					      ep0_state_strings[ep0_state],\
-					      ep0_state_strings[s],	\
-					      __PRETTY_FUNCTION__,	\
-					      __LINE__);		\
-			ep0_state = s;					\
-		}							\
-	} else {							\
-		if (debug_level > 0)					\
-			serial_printf("Error at %s %d with setting "	\
-				      "state %d is invalid\n",		\
-				      __PRETTY_FUNCTION__, __LINE__, s); \
-	}								\
-} while (0)
-
-/* static implies these initialized to 0 or NULL */
-static int debug_setup;
-static int debug_level;
-static struct musb_epinfo epinfo[MAX_ENDPOINT * 2];
-static enum ep0_state_enum {
-	IDLE = 0,
-	TX,
-	RX,
-	SET_ADDRESS
-} ep0_state = IDLE;
-static char *ep0_state_strings[4] = {
-	"IDLE",
-	"TX",
-	"RX",
-	"SET_ADDRESS",
-};
-
-static struct urb *ep0_urb;
-struct usb_endpoint_instance *ep0_endpoint;
-static struct usb_device_instance *udc_device;
-static int enabled;
-
-#ifdef MUSB_DEBUG
-static void musb_db_regs(void)
-{
-	u8 b;
-	u16 w;
-
-	b = readb(&musbr->faddr);
-	serial_printf("\tfaddr   0x%2.2x\n", b);
-
-	b = readb(&musbr->power);
-	musb_print_pwr(b);
-
-	w = readw(&musbr->ep[0].ep0.csr0);
-	musb_print_csr0(w);
-
-	b = readb(&musbr->devctl);
-	musb_print_devctl(b);
-
-	b = readb(&musbr->ep[0].ep0.configdata);
-	musb_print_config(b);
-
-	w = readw(&musbr->frame);
-	serial_printf("\tframe   0x%4.4x\n", w);
-
-	b = readb(&musbr->index);
-	serial_printf("\tindex   0x%2.2x\n", b);
-
-	w = readw(&musbr->ep[1].epN.rxmaxp);
-	musb_print_rxmaxp(w);
-
-	w = readw(&musbr->ep[1].epN.rxcsr);
-	musb_print_rxcsr(w);
-
-	w = readw(&musbr->ep[1].epN.txmaxp);
-	musb_print_txmaxp(w);
-
-	w = readw(&musbr->ep[1].epN.txcsr);
-	musb_print_txcsr(w);
-}
-#else
-#define musb_db_regs()
-#endif /* DEBUG_MUSB */
-
-static void musb_peri_softconnect(void)
-{
-	u8 power, devctl;
-	u8 intrusb;
-	u16 intrrx, intrtx;
-
-	/* Power off MUSB */
-	power = readb(&musbr->power);
-	power &= ~MUSB_POWER_SOFTCONN;
-	writeb(power, &musbr->power);
-
-	/* Read intr to clear */
-	intrusb = readb(&musbr->intrusb);
-	intrrx = readw(&musbr->intrrx);
-	intrtx = readw(&musbr->intrtx);
-
-	udelay(1000 * 1000); /* 1 sec */
-
-	/* Power on MUSB */
-	power = readb(&musbr->power);
-	power |= MUSB_POWER_SOFTCONN;
-	/*
-	 * The usb device interface is usb 1.1
-	 * Disable 2.0 high speed by clearring the hsenable bit.
-	 */
-	power &= ~MUSB_POWER_HSENAB;
-	writeb(power, &musbr->power);
-
-	/* Check if device is in b-peripheral mode */
-	devctl = readb(&musbr->devctl);
-	if (!(devctl & MUSB_DEVCTL_BDEVICE) ||
-	    (devctl & MUSB_DEVCTL_HM)) {
-		serial_printf("ERROR : Unsupport USB mode\n");
-		serial_printf("Check that mini-B USB cable is attached "
-			      "to the device\n");
-	}
-
-	if (debug_setup && (debug_level > 1))
-		musb_db_regs();
-}
-
-static void musb_peri_reset(void)
-{
-	if ((debug_setup) && (debug_level > 1))
-		serial_printf("INFO : %s reset\n", __PRETTY_FUNCTION__);
+#define UDC_BASE MENTOR_USB0_BASE
 
-	if (ep0_endpoint)
-		ep0_endpoint->endpoint_address = 0xff;
+/* Private definitions */
+enum ep0_status { IDLE, DATA_STAGE, DATA_COMPLETE };
 
-	/* Sync sw and hw addresses */
-	writeb(udc_device->address, &musbr->faddr);
-
-	SET_EP0_STATE(IDLE);
-}
-
-static void musb_peri_resume(void)
-{
-	/* noop */
-}
-
-static void musb_peri_ep0_stall(void)
-{
-	u16 csr0;
-
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-	csr0 |= MUSB_CSR0_P_SENDSTALL;
-	writew(csr0, &musbr->ep[0].ep0.csr0);
-	if ((debug_setup) && (debug_level > 1))
-		serial_printf("INFO : %s stall\n", __PRETTY_FUNCTION__);
-}
+/* Private variables */
+static struct usb_device_instance *udc_device;
+static enum ep0_status ep0status = IDLE;
+static unsigned char do_set_address = 0;
+static struct urb *ep0_urb = NULL;
 
-static void musb_peri_ep0_ack_req(void)
+/* Helper functions */
+static void insl(u32 reg, u32 *data, u32 size)
 {
-	u16 csr0;
+	u32 t;
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-	csr0 |= MUSB_CSR0_P_SVDRXPKTRDY;
-	writew(csr0, &musbr->ep[0].ep0.csr0);
+	for (t = 0; t < size; t++, data++)
+		*data = readl(reg);
 }
 
-static void musb_ep0_tx_ready(void)
+static void outsl(u32 reg, u32 *data, u32 size)
 {
-	u16 csr0;
+	u32 t;
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-	csr0 |= MUSB_CSR0_TXPKTRDY;
-	writew(csr0, &musbr->ep[0].ep0.csr0);
+	for (t = 0; t < size; t++, data++)
+		writel(*data, reg);
 }
 
-static void musb_ep0_tx_ready_and_last(void)
+static void outsb(u32 reg, u8 *data, u32 size)
 {
-	u16 csr0;
+	u32 t;
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-	csr0 |= (MUSB_CSR0_TXPKTRDY | MUSB_CSR0_P_DATAEND);
-	writew(csr0, &musbr->ep[0].ep0.csr0);
+	for (t = 0; t < size; t++, data++)
+		writeb(*data, reg);
 }
 
-static void musb_peri_ep0_last(void)
+static void musb_fifo_read(int epnumber, u8 *data, u32 size)
 {
-	u16 csr0;
+	if ((u32)data & 0x3) {		/* Not aligned data */
+		insb((UDC_FIFO0 + (epnumber << 2)), data, size);
+	} else {			/* 32 bits aligned data */
+		int i;
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-	csr0 |= MUSB_CSR0_P_DATAEND;
-	writew(csr0, &musbr->ep[0].ep0.csr0);
-}
-
-static void musb_peri_ep0_set_address(void)
-{
-	u8 faddr;
-	writeb(udc_device->address, &musbr->faddr);
-
-	/* Verify */
-	faddr = readb(&musbr->faddr);
-	if (udc_device->address == faddr) {
-		SET_EP0_STATE(IDLE);
-		usbd_device_event_irq(udc_device, DEVICE_ADDRESS_ASSIGNED, 0);
-		if ((debug_setup) && (debug_level > 1))
-			serial_printf("INFO : %s Address set to %d\n",
-				      __PRETTY_FUNCTION__, udc_device->address);
-	} else {
-		if (debug_level > 0)
-			serial_printf("ERROR : %s Address missmatch "
-				      "sw %d vs hw %d\n",
-				      __PRETTY_FUNCTION__,
-				      udc_device->address, faddr);
+		insl(UDC_FIFO0 + (epnumber << 2), (u32 *)data, size >> 2);
+		data += size & ~0x3;
+		i = size & 0x3;
+		while (i--) {
+			*data = readb(UDC_FIFO0 + (epnumber << 2));
+			data++;
+		}
 	}
 }
 
-static void musb_peri_rx_ack(unsigned int ep)
+static void musb_fifo_write(int epnumber, u8 *data, u32 size)
 {
-	u16 peri_rxcsr;
+	if ((u32)data & 0x3) {		/* Not aligned data */
+		outsb(UDC_FIFO0 + (epnumber << 2), data, size);
+	} else {			/* 32 bits aligned data */
+		int i;
 
-	peri_rxcsr = readw(&musbr->ep[ep].epN.rxcsr);
-	peri_rxcsr &= ~MUSB_RXCSR_RXPKTRDY;
-	writew(peri_rxcsr, &musbr->ep[ep].epN.rxcsr);
+		outsl(UDC_FIFO0 + (epnumber << 2), (u32 *)data, size >> 2);
+		data += size & ~0x3;
+		i = size & 0x3;
+		while (i--) {
+			writeb(*data, UDC_FIFO0 + (epnumber << 2));
+			data++;
+		}
+	}
 }
 
-static void musb_peri_tx_ready(unsigned int ep)
+static void musb_fifos_configure(struct usb_device_instance *device)
 {
-	u16 peri_txcsr;
+	int ep;
+	struct usb_bus_instance *bus;
+	struct usb_endpoint_instance *endpoint;
+	unsigned short ep_ptr, ep_size, ep_doublebuffer;
+	int ep_addr, packet_size, buffer_size, attributes;
 
-	peri_txcsr = readw(&musbr->ep[ep].epN.txcsr);
-	peri_txcsr |= MUSB_TXCSR_TXPKTRDY;
-	writew(peri_txcsr, &musbr->ep[ep].epN.txcsr);
-}
+	bus = device->bus;
 
-static void musb_peri_ep0_zero_data_request(int err)
-{
-	musb_peri_ep0_ack_req();
+	ep_ptr = 0;
 
-	if (err) {
-		musb_peri_ep0_stall();
-		SET_EP0_STATE(IDLE);
-	} else {
-
-		musb_peri_ep0_last();
+	for (ep = 0; ep < bus->max_endpoints; ep++) {
+		endpoint = bus->endpoint_array + ep;
+		ep_addr = endpoint->endpoint_address;
+		if ((ep_addr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
+			/* IN endpoint */
+			packet_size = endpoint->tx_packetSize;
+			attributes = endpoint->tx_attributes;
+		} else {
+			/* OUT endpoint */
+			packet_size = endpoint->rcv_packetSize;
+			attributes = endpoint->rcv_attributes;
+		}
 
-		/* USBD state */
-		switch (ep0_urb->device_request.bRequest) {
-		case USB_REQ_SET_ADDRESS:
-			if ((debug_setup) && (debug_level > 1))
-				serial_printf("INFO : %s received set "
-					      "address\n", __PRETTY_FUNCTION__);
+		switch (packet_size) {
+		case 0:
+			ep_size = 0;
 			break;
-
-		case USB_REQ_SET_CONFIGURATION:
-			if ((debug_setup) && (debug_level > 1))
-				serial_printf("INFO : %s Configured\n",
-					      __PRETTY_FUNCTION__);
-			usbd_device_event_irq(udc_device, DEVICE_CONFIGURED, 0);
+		case 8:
+			ep_size = 0;
+			break;
+		case 16:
+			ep_size = 1;
+			break;
+		case 32:
+			ep_size = 2;
+			break;
+		case 64:
+			ep_size = 3;
+			break;
+		case 128:
+			ep_size = 4;
+			break;
+		case 256:
+			ep_size = 5;
+			break;
+		case 512:
+			ep_size = 6;
+			break;
+		default:
+			printf("ep 0x%02x has bad packet size %d",
+				ep_addr, packet_size);
+			packet_size = 0;
+			ep_size = 0;
 			break;
 		}
 
-		/* EP0 state */
-		if (USB_REQ_SET_ADDRESS == ep0_urb->device_request.bRequest) {
-			SET_EP0_STATE(SET_ADDRESS);
-		} else {
-			SET_EP0_STATE(IDLE);
+		switch (attributes & USB_ENDPOINT_XFERTYPE_MASK) {
+		case USB_ENDPOINT_XFER_CONTROL:
+		case USB_ENDPOINT_XFER_BULK:
+		case USB_ENDPOINT_XFER_INT:
+		default:
+			/* A non-isochronous endpoint may optionally be
+			 * double-buffered. For now we disable
+			 * double-buffering.
+			 */
+			ep_doublebuffer = 0;
+			if (packet_size > 64)
+				packet_size = 0;
+			if (!ep || !ep_doublebuffer)
+				buffer_size = packet_size;
+			else
+				buffer_size = packet_size * 2;
+			break;
+		case USB_ENDPOINT_XFER_ISOC:
+			/* Isochronous endpoints are always double-
+			 * buffered
+			 */
+			ep_doublebuffer = 1;
+			buffer_size = packet_size * 2;
+			break;
 		}
-	}
-}
-
-static void musb_peri_ep0_rx_data_request(void)
-{
-	/*
-	 * This is the completion of the data OUT / RX
-	 *
-	 * Host is sending data to ep0 that is not
-	 * part of setup.  This comes from the cdc_recv_setup
-	 * op that is device specific.
-	 *
-	 */
-	musb_peri_ep0_ack_req();
 
-	ep0_endpoint->rcv_urb = ep0_urb;
-	ep0_urb->actual_length = 0;
-	SET_EP0_STATE(RX);
-}
+		/* check to see if our packet buffer RAM is exhausted */
+		if ((ep_ptr + buffer_size) > UDC_MAX_FIFO_SIZE) {
+			printf("out of packet RAM for ep 0x%02x buf size %d",
+				ep_addr, buffer_size);
+			buffer_size = packet_size = 0;
+		}
 
-static void musb_peri_ep0_tx_data_request(int err)
-{
-	if (err) {
-		musb_peri_ep0_stall();
-		SET_EP0_STATE(IDLE);
-	} else {
-		musb_peri_ep0_ack_req();
+		/* force a default configuration for endpoint 0 since it is
+		 * always enabled
+		 */
+		if (!ep && ((packet_size < 8) || (packet_size > 64))) {
+			buffer_size = packet_size = 64;
+			ep_size = 3;
+		}
 
-		ep0_endpoint->tx_urb = ep0_urb;
-		ep0_endpoint->sent = 0;
-		SET_EP0_STATE(TX);
+		writeb(ep & 0xF, UDC_INDEX);
+		if ((ep_addr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
+			/* IN endpoint */
+			writeb((ep_doublebuffer << 4) | (ep_size & 0xf),
+				UDC_TXFIFOSZ);
+			writew(ep_ptr >> 3, UDC_TXFIFOADDR);
+			if (!ep) {	/* This only apply for ep != 0 */
+				writew(packet_size & 0x3FF, UDC_TXMAXP);
+			}
+		} else {
+			/* OUT endpoint */
+			writeb((ep_doublebuffer << 4) | (ep_size & 0xf),
+				UDC_RXFIFOSZ);
+			writew(ep_ptr >> 3, UDC_RXFIFOADDR);
+			if (!ep) {	/* This only apply for ep != 0 */
+				writew(packet_size & 0x3FF, UDC_RXMAXP);
+			}
+		}
+		ep_ptr += buffer_size;
 	}
 }
 
-static void musb_peri_ep0_idle(void)
+static void musb_ep0_tx(struct usb_endpoint_instance *endpoint)
 {
-	u16 count0;
-	int err;
-	u16 csr0;
-
-	/*
-	 * Verify addresses
-	 * A lot of confusion can be caused if the address
-	 * in software, udc layer, does not agree with the
-	 * hardware.  Since the setting of the hardware address
-	 * must be set after the set address request, the
-	 * usb state machine is out of sync for a few frame.
-	 * It is a good idea to run this check when changes
-	 * are made to the state machine.
-	 */
-	if ((debug_level > 0) &&
-	    (ep0_state != SET_ADDRESS)) {
-		u8 faddr;
-
-		faddr = readb(&musbr->faddr);
-		if (udc_device->address != faddr) {
-			serial_printf("ERROR : %s addresses do not"
-				      "match sw %d vs hw %d\n",
-				      __PRETTY_FUNCTION__,
-				      udc_device->address, faddr);
-			udelay(1000 * 1000);
-			hang();
-		}
-	}
+	unsigned int size = 0;
+	struct urb *urb = endpoint->tx_urb;
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
+	writeb(0, UDC_INDEX);
 
-	if (!(MUSB_CSR0_RXPKTRDY & csr0))
-		goto end;
+	if (urb) {
 
-	count0 = readw(&musbr->ep[0].ep0.count0);
-	if (count0 == 0)
-		goto end;
+		if ((size =
+			MIN(urb->actual_length - endpoint->sent,
+				endpoint->tx_packetSize))) {
 
-	if (count0 != 8) {
-		if ((debug_setup) && (debug_level > 1))
-			serial_printf("WARN : %s SETUP incorrect size %d\n",
-				      __PRETTY_FUNCTION__, count0);
-		musb_peri_ep0_stall();
-		goto end;
-	}
-
-	read_fifo(0, count0, &ep0_urb->device_request);
-
-	if (debug_level > 2)
-		print_usb_device_request(&ep0_urb->device_request);
-
-	if (ep0_urb->device_request.wLength == 0) {
-		err = ep0_recv_setup(ep0_urb);
-
-		/* Zero data request */
-		musb_peri_ep0_zero_data_request(err);
-	} else {
-		/* Is data coming or going ? */
-		u8 reqType = ep0_urb->device_request.bmRequestType;
+			musb_fifo_write(0, urb->buffer + endpoint->sent, size);
+		}
+		endpoint->last = size;
 
-		if (USB_REQ_DEVICE2HOST == (reqType & USB_REQ_DIRECTION_MASK)) {
-			err = ep0_recv_setup(ep0_urb);
-			/* Device to host */
-			musb_peri_ep0_tx_data_request(err);
+		if (((endpoint->sent + size) == ep0_urb->device_request.wLength)
+			|| (size != endpoint->tx_packetSize)) {
+			ep0status = DATA_COMPLETE;
+			/* Transmit packet and set data end */
+			writew(0xA, UDC_CSR0);
 		} else {
-			/*
-			 * Host to device
-			 *
-			 * The RX routine will call ep0_recv_setup
-			 * when the data packet has arrived.
-			 */
-			musb_peri_ep0_rx_data_request();
+			writew(0x2, UDC_CSR0);	/* Transmit packet */
 		}
 	}
-
-end:
-	return;
 }
 
-static void musb_peri_ep0_rx(void)
+static void musb_ep0_handler(struct usb_endpoint_instance *endpoint)
 {
-	/*
-	 * This is the completion of the data OUT / RX
-	 *
-	 * Host is sending data to ep0 that is not
-	 * part of setup.  This comes from the cdc_recv_setup
-	 * op that is device specific.
-	 *
-	 * Pass the data back to driver ep0_recv_setup which
-	 * should give the cdc_recv_setup the chance to handle
-	 * the rx
-	 */
 	u16 csr0;
-	u16 count0;
 
-	if (debug_level > 3) {
-		if (0 != ep0_urb->actual_length) {
-			serial_printf("%s finished ? %d of %d\n",
-				      __PRETTY_FUNCTION__,
-				      ep0_urb->actual_length,
-				      ep0_urb->device_request.wLength);
-		}
-	}
+	writeb(0, UDC_INDEX);
 
-	if (ep0_urb->device_request.wLength == ep0_urb->actual_length) {
-		musb_peri_ep0_last();
-		SET_EP0_STATE(IDLE);
-		ep0_recv_setup(ep0_urb);
-		return;
+	/* Check errors */
+	csr0 = readw(UDC_CSR0);
+
+	if (csr0 & 0x4) {			/* Sent stall */
+		writew(csr0 & ~0x4, UDC_CSR0);	/* Clear stall */
+		printf("%s: stall received on EP0!\n", __FUNCTION__);
 	}
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-	if (!(MUSB_CSR0_RXPKTRDY & csr0))
+	if (csr0 & 0x10) {		/* Setup end */
+		writew(0x80, UDC_CSR0);	/* Clear setup end */
+		printf("%s: setup END early happened! status is %d\n",
+			__FUNCTION__, ep0status);
+		ep0status = IDLE;
 		return;
+	}
 
-	count0 = readw(&musbr->ep[0].ep0.count0);
-
-	if (count0) {
-		struct usb_endpoint_instance *endpoint;
-		u32 length;
-		u8 *data;
-
-		endpoint = ep0_endpoint;
-		if (endpoint && endpoint->rcv_urb) {
-			struct urb *urb = endpoint->rcv_urb;
-			unsigned int remaining_space = urb->buffer_length -
-				urb->actual_length;
-
-			if (remaining_space) {
-				int urb_bad = 0; /* urb is good */
-
-				if (count0 > remaining_space)
-					length = remaining_space;
-				else
-					length = count0;
-
-				data = (u8 *) urb->buffer_data;
-				data += urb->actual_length;
-
-				/* The common musb fifo reader */
-				read_fifo(0, length, data);
-
-				musb_peri_ep0_ack_req();
-
+	switch (ep0status) {
+	case DATA_COMPLETE:
+		if (do_set_address) {
+			/*
+			 * We need to set the address only after
+			 * the status stage is complete
+			 */
+			writeb(udc_device->address, UDC_FADDR);
+			do_set_address = 0;
+		}
+		ep0status = IDLE;
+		/* Fallthrough */
+	case IDLE:		/* Receiving a setup packet */
+		if (csr0 & 0x1) {
+			insl(UDC_FIFO0,
+				(unsigned int *) &ep0_urb->device_request, 2);
+
+			/* If we have data, then don't go to IDLE state */
+			if (ep0_urb->device_request.wLength) {
+				ep0status = DATA_STAGE;
+
+				writew(0x40, UDC_CSR0);	/* Clear RXPKTRDY */
+				if ((ep0_urb->device_request.
+					bmRequestType & USB_REQ_DIRECTION_MASK)
+					== USB_REQ_DEVICE2HOST) {
+
+					/* Try to process setup packet */
+					if (ep0_recv_setup(ep0_urb)) {
+						/*
+						 * Not a setup packet, stall
+						 * next EP0 transaction
+						 */
+						writew(0x20, UDC_CSR0);
+						ep0status = IDLE;
+						return;
+					}
+					/*
+					 * If we are sending data, do it now, as
+					 * ep0_recv_setup should have prepare
+					 * them
+					 */
+					endpoint->tx_urb = ep0_urb;
+					endpoint->sent = 0;
+
+					musb_ep0_tx(endpoint);
+				} else {
+					endpoint->rcv_urb = ep0_urb;
+					ep0_urb->actual_length = 0;
+				}
+			} else {	/* Processing zero-length packet */
 				/*
-				 * urb's actual_length is updated in
-				 * usbd_rcv_complete
+				 * The www.linux-usb.org/usbtest 'test 14'
+				 * fails with error for zero length request.
+				 * If the SETUP packet requests ZERO length data
+				 * from device-to-host, the TXPKTRDY bit needs
+				 * to be set in TXCSR otherwise the STATUS stage
+				 * of control transfer will never complete.
 				 */
-				usbd_rcv_complete(endpoint, length, urb_bad);
+				if ((ep0_urb->device_request.
+					bmRequestType & USB_REQ_DIRECTION_MASK)
+					== USB_REQ_DEVICE2HOST) {
+					/*
+					 * Clear RXPKTRDY and DATAEND and
+					 * TXPKTRDY
+					 */
+					writew(0x4A, UDC_CSR0);
+				} else {
+					/* Clear RXPKTRDY and DATAEND */
+					writew(0x48, UDC_CSR0);
+				}
+
+				/* Try to process setup packet */
+				if (ep0_recv_setup(ep0_urb)) {
+					/*
+					 * Not a setup packet, stall next EP0
+					 * transaction
+					 */
+					writew(0x20, UDC_CSR0);
+					ep0status = IDLE;
+					return;
+				}
+
+				switch (ep0_urb->device_request.bRequest) {
+				case USB_REQ_SET_ADDRESS:
+					usbd_device_event_irq(udc_device,
+						DEVICE_ADDRESS_ASSIGNED, 0);
+					do_set_address = 1;
+					break;
+				case USB_REQ_SET_CONFIGURATION:
+					usbd_device_event_irq(udc_device,
+						DEVICE_CONFIGURED, 0);
+					break;
+				}
 
-			} else {
-				if (debug_level > 0)
-					serial_printf("ERROR : %s no space in "
-						      "rcv buffer\n",
-						      __PRETTY_FUNCTION__);
+				ep0status = DATA_COMPLETE;
+			}
+		}
+		break;
+	case DATA_STAGE:
+		if ((ep0_urb->device_request.
+			bmRequestType & USB_REQ_DIRECTION_MASK)
+			== USB_REQ_DEVICE2HOST) {
+			if (!(csr0 & 0x2)) {	/* There packet was send? */
+				endpoint->sent += endpoint->last;
+				/*
+				 * If we finished sending data we would not
+				 * be on the DATA_STAGE
+				 */
+				musb_ep0_tx(endpoint);
 			}
 		} else {
-			if (debug_level > 0)
-				serial_printf("ERROR : %s problem with "
-					      "endpoint\n",
-					      __PRETTY_FUNCTION__);
+			/* Receiving data */
+			u16 length = readw(UDC_COUNT0);
+
+			if (length) {
+				if (ep0_urb->actual_length + length >
+					ep0_urb->device_request.wLength)
+					length =
+					ep0_urb->device_request.wLength -
+					ep0_urb->actual_length;
+
+				endpoint->last = length;
+
+				musb_fifo_read(0, &ep0_urb->
+					buffer[ep0_urb->actual_length], length);
+				ep0_urb->actual_length += length;
+			}
+
+			/*
+			 * We finish if we received the amount of data expected,
+			 * or less of the packet size
+			 */
+			if ((ep0_urb->actual_length ==
+				ep0_urb->device_request.wLength) ||
+				(endpoint->last != endpoint->tx_packetSize)) {
+				ep0status = DATA_COMPLETE;
+				/* Clear RXPKTRDY and DATAEND */
+				writew(0x48, UDC_CSR0);
+				/* This will process the incoming data */
+                                //printf("musb %08x %08x\n", ((int *)ep0_urb->buffer)[0], ((int *)ep0_urb->buffer)[1]);
+				if (ep0_recv_setup(ep0_urb)) {
+					/*
+					 * Not a setup packet, stall next EP0
+					 * transaction
+					 */
+					writew(0x20, UDC_CSR0);
+					return;
+				}
+			} else
+				writew(0x40, UDC_CSR0);	/* Clear RXPKTRDY */
 		}
-	} else {
-		if (debug_level > 0)
-			serial_printf("ERROR : %s with nothing to do\n",
-				      __PRETTY_FUNCTION__);
+		break;
 	}
 }
 
-static void musb_peri_ep0_tx(void)
+static void musb_ep_tx(struct usb_endpoint_instance *endpoint)
 {
-	u16 csr0;
-	int transfer_size = 0;
-	unsigned int p, pm;
-
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
-
-	/* Check for pending tx */
-	if (csr0 & MUSB_CSR0_TXPKTRDY)
-		goto end;
-
-	/* Check if this is the last packet sent */
-	if (ep0_endpoint->sent >= ep0_urb->actual_length) {
-		SET_EP0_STATE(IDLE);
-		goto end;
-	}
-
-	transfer_size = ep0_urb->actual_length - ep0_endpoint->sent;
-	/* Is the transfer size negative ? */
-	if (transfer_size <= 0) {
-		if (debug_level > 0)
-			serial_printf("ERROR : %s problem with the"
-				      " transfer size %d\n",
-				      __PRETTY_FUNCTION__,
-				      transfer_size);
-		SET_EP0_STATE(IDLE);
-		goto end;
-	}
-
-	/* Truncate large transfers to the fifo size */
-	if (transfer_size > ep0_endpoint->tx_packetSize)
-		transfer_size = ep0_endpoint->tx_packetSize;
+	unsigned int size = 0, epnumber =
+		endpoint->endpoint_address & USB_ENDPOINT_NUMBER_MASK;
+	struct urb *urb = endpoint->tx_urb;
 
-	write_fifo(0, transfer_size, &ep0_urb->buffer[ep0_endpoint->sent]);
-	ep0_endpoint->sent += transfer_size;
-
-	/* Done or more to send ? */
-	if (ep0_endpoint->sent >= ep0_urb->actual_length)
-		musb_ep0_tx_ready_and_last();
-	else
-		musb_ep0_tx_ready();
-
-	/* Wait a bit */
-	pm = 10;
-	for (p = 0; p < pm; p++) {
-		csr0 = readw(&musbr->ep[0].ep0.csr0);
-		if (!(csr0 & MUSB_CSR0_TXPKTRDY))
-			break;
+	writeb(epnumber, UDC_INDEX);
+	if (urb) {
+		if ((size =
+			MIN(urb->actual_length - endpoint->sent,
+			endpoint->tx_packetSize))) {
+			musb_fifo_write(epnumber, urb->buffer + endpoint->sent,
+					size);
+		}
+		endpoint->last = size;
+		endpoint->state = 1; /* Transmit hardware is busy */
 
-		/* Double the delay. */
-		udelay(1 << pm);
+		writew(readw(UDC_TXCSR) | 0x1, UDC_TXCSR); /* Transmit packet */
 	}
-
-	if ((ep0_endpoint->sent >= ep0_urb->actual_length) && (p < pm))
-		SET_EP0_STATE(IDLE);
-
-end:
-	return;
 }
 
-static void musb_peri_ep0(void)
+
+static void musb_tx_handler(struct usb_endpoint_instance *endpoint)
 {
-	u16 csr0;
+	unsigned int epnumber =
+		endpoint->endpoint_address & USB_ENDPOINT_NUMBER_MASK;
+	u16 txcsr;
 
-	if (SET_ADDRESS == ep0_state)
-		return;
+	writeb(epnumber, UDC_INDEX);
 
-	csr0 = readw(&musbr->ep[0].ep0.csr0);
+	/* Check errors */
+	txcsr = readw(UDC_TXCSR);
 
-	/* Error conditions */
-	if (MUSB_CSR0_P_SENTSTALL & csr0) {
-		csr0 &= ~MUSB_CSR0_P_SENTSTALL;
-		writew(csr0, &musbr->ep[0].ep0.csr0);
-		SET_EP0_STATE(IDLE);
+	if (txcsr & 0x4) {		/* Clear underrun */
+		txcsr &= ~0x4;
 	}
-	if (MUSB_CSR0_P_SETUPEND & csr0) {
-		csr0 |= MUSB_CSR0_P_SVDSETUPEND;
-		writew(csr0, &musbr->ep[0].ep0.csr0);
-		SET_EP0_STATE(IDLE);
-		if ((debug_setup) && (debug_level > 1))
-			serial_printf("WARN: %s SETUPEND\n",
-				      __PRETTY_FUNCTION__);
+	if (txcsr & 0x20) {		/* SENTSTALL */
+		writew(txcsr & ~0x20, UDC_TXCSR);	/* Clear stall */
+		return;
 	}
 
-	/* Normal states */
-	if (IDLE == ep0_state)
-		musb_peri_ep0_idle();
-
-	if (TX == ep0_state)
-		musb_peri_ep0_tx();
+	if (endpoint->tx_urb && !(txcsr & 0x1)) { /* The packet was send? */
+		if ((endpoint->sent + endpoint->last == endpoint->tx_urb->
+			actual_length)	/* Send a zero length packet? */
+			&& (endpoint->last == endpoint->tx_packetSize)) {
+			/* Prepare to transmit a zero-length packet. */
+			endpoint->sent += endpoint->last;
+			musb_ep_tx(endpoint);
+		} else if (endpoint->tx_urb->actual_length) {
+			/* retire the data that was just sent */
+			usbd_tx_complete(endpoint);
+			endpoint->state = 0; /* Transmit hardware is free */
 
-	if (RX == ep0_state)
-		musb_peri_ep0_rx();
+			/*
+			 * Check to see if we have more data ready to transmit
+			 * now.
+			 */
+			if (endpoint->tx_urb && endpoint->tx_urb->
+				actual_length) {
+				musb_ep_tx(endpoint);
+			}
+		}
+	}
 }
 
-static void musb_peri_rx_ep(unsigned int ep)
+static void musb_rx_handler(struct usb_endpoint_instance *endpoint)
 {
-	u16 peri_rxcount = readw(&musbr->ep[ep].epN.rxcount);
+	unsigned int epnumber =
+		endpoint->endpoint_address & USB_ENDPOINT_NUMBER_MASK;
+	u16 rxcsr;
+	u16 length;
 
-	if (peri_rxcount) {
-		struct usb_endpoint_instance *endpoint;
-		u32 length;
-		u8 *data;
+	writeb(epnumber, UDC_INDEX);
 
-		endpoint = GET_ENDPOINT(udc_device, ep);
-		if (endpoint && endpoint->rcv_urb) {
-			struct urb *urb = endpoint->rcv_urb;
-			unsigned int remaining_space = urb->buffer_length -
-				urb->actual_length;
+	/* Check errors */
+	rxcsr = readw(UDC_RXCSR);
 
-			if (remaining_space) {
-				int urb_bad = 0; /* urb is good */
-
-				if (peri_rxcount > remaining_space)
-					length = remaining_space;
-				else
-					length = peri_rxcount;
-
-				data = (u8 *) urb->buffer_data;
-				data += urb->actual_length;
-
-				/* The common musb fifo reader */
-				read_fifo(ep, length, data);
+	if (!(rxcsr & 0x1))		/* There is a package received? */
+		return;
 
-				musb_peri_rx_ack(ep);
+	if (rxcsr & 0x40)		/* SENTSTALL */
+		writew(rxcsr & ~0x40, UDC_RXCSR);	/* Clear stall */
 
-				/*
-				 * urb's actual_length is updated in
-				 * usbd_rcv_complete
-				 */
-				usbd_rcv_complete(endpoint, length, urb_bad);
+	length = readw(UDC_RXCOUNT);
 
-			} else {
-				if (debug_level > 0)
-					serial_printf("ERROR : %s %d no space "
-						      "in rcv buffer\n",
-						      __PRETTY_FUNCTION__, ep);
-			}
-		} else {
-			if (debug_level > 0)
-				serial_printf("ERROR : %s %d problem with "
-					      "endpoint\n",
-					      __PRETTY_FUNCTION__, ep);
+	if (endpoint->rcv_urb) {
+		/* Receiving data */
+		if (length) {
+			musb_fifo_read(epnumber, &endpoint->rcv_urb->
+				buffer[endpoint->rcv_urb->actual_length],
+				length);
+			writew(rxcsr & ~0x1, UDC_RXCSR);	/* Clear RXPKTRDY */
+			usbd_rcv_complete(endpoint, length, 0);
 		}
-
 	} else {
-		if (debug_level > 0)
-			serial_printf("ERROR : %s %d with nothing to do\n",
-				      __PRETTY_FUNCTION__, ep);
+		printf("%s: no receive URB!\n", __FUNCTION__);
 	}
 }
 
-static void musb_peri_rx(u16 intr)
+static void musb_reset(void)
 {
-	unsigned int ep;
-
-	/* Check for EP0 */
-	if (0x01 & intr)
-		musb_peri_ep0();
-
-	for (ep = 1; ep < 16; ep++) {
-		if ((1 << ep) & intr)
-			musb_peri_rx_ep(ep);
-	}
-}
-
-static void musb_peri_tx(u16 intr)
-{
-	/* Check for EP0 */
-	if (0x01 & intr)
-		musb_peri_ep0_tx();
-
-	/*
-	 * Use this in the future when handling epN tx
-	 *
-	 * u8 ep;
-	 *
-	 * for (ep = 1; ep < 16; ep++) {
-	 *	if ((1 << ep) & intr) {
-	 *		/ * handle tx for this endpoint * /
-	 *	}
-	 * }
-	 */
+	usbd_device_event_irq(udc_device, DEVICE_HUB_CONFIGURED, 0);
+	usbd_device_event_irq(udc_device, DEVICE_RESET, 0);
+	ep0status = IDLE;
+	do_set_address = 0;
 }
 
+/* Public functions - called by usbdcore, usbtty, etc. */
 void udc_irq(void)
 {
-	/* This is a high freq called function */
-	if (enabled) {
-		u8 intrusb;
-
-		intrusb = readb(&musbr->intrusb);
-
-		/*
-		 * See drivers/usb/gadget/mpc8xx_udc.c for
-		 * state diagram going from detached through
-		 * configuration.
-		 */
-		if (MUSB_INTR_RESUME & intrusb) {
-			usbd_device_event_irq(udc_device,
-					      DEVICE_BUS_ACTIVITY, 0);
-			musb_peri_resume();
-		}
-
-		musb_peri_ep0();
-
-		if (MUSB_INTR_RESET & intrusb) {
-			usbd_device_event_irq(udc_device, DEVICE_RESET, 0);
-			musb_peri_reset();
+	unsigned char int_usb = readb(UDC_INTRUSB);
+	unsigned short int_tx = readw(UDC_INTRTX);
+	unsigned short int_rx = readw(UDC_INTRRX);
+	int ep;
+
+	if (int_usb) {
+		if (int_usb & 0x4) {	/* Reset */
+			/* The controller clears FADDR, INDEX, and FIFOs */
+			musb_reset();
 		}
-
-		if (MUSB_INTR_DISCONNECT & intrusb) {
-			/* cable unplugged from hub/host */
-			usbd_device_event_irq(udc_device, DEVICE_RESET, 0);
-			musb_peri_reset();
+		if (int_usb & 0x20) {	/* Disconnected */
 			usbd_device_event_irq(udc_device, DEVICE_HUB_RESET, 0);
 		}
-
-		if (MUSB_INTR_SOF & intrusb) {
+		if (int_usb & 0x1)
 			usbd_device_event_irq(udc_device,
-					      DEVICE_BUS_ACTIVITY, 0);
-			musb_peri_resume();
-		}
-
-		if (MUSB_INTR_SUSPEND & intrusb) {
+					DEVICE_BUS_INACTIVE, 0);
+		if (int_usb & 0x2)
 			usbd_device_event_irq(udc_device,
-					      DEVICE_BUS_INACTIVE, 0);
-		}
-
-		if (ep0_state != SET_ADDRESS) {
-			u16 intrrx, intrtx;
-
-			intrrx = readw(&musbr->intrrx);
-			intrtx = readw(&musbr->intrtx);
+					DEVICE_BUS_ACTIVITY, 0);
+	}
 
-			if (intrrx)
-				musb_peri_rx(intrrx);
+	/* Note: IRQ values auto clear so read just before processing */
+	if (int_rx) {		/* OUT endpoints */
+		ep = 1;
+		int_rx >>= 1;
+		while (int_rx) {
+			if (int_rx & 1)
+				musb_rx_handler(udc_device->bus->endpoint_array
+						+ ep);
+			int_rx >>= 1;
+			ep++;
+		}
+	}
+	if (int_tx) {		/* IN endpoints */
+		if (int_tx & 1)
+			musb_ep0_handler(udc_device->bus->endpoint_array);
 
-			if (intrtx)
-				musb_peri_tx(intrtx);
-		} else {
-			if (MUSB_INTR_SOF & intrusb) {
-				u8 faddr;
-				faddr = readb(&musbr->faddr);
-				/*
-				 * Setting of the address can fail.
-				 * Normally it succeeds the second time.
-				 */
-				if (udc_device->address != faddr)
-					musb_peri_ep0_set_address();
-			}
+		ep = 1;
+		int_tx >>= 1;
+		while (int_tx) {
+			if (int_tx & 1)
+				musb_tx_handler(udc_device->bus->endpoint_array
+						+ ep);
+			int_tx >>= 1;
+			ep++;
 		}
 	}
 }
 
-void udc_set_nak(int ep_num)
+/* Turn on the USB connection */
+void udc_connect(void)
 {
-	/* noop */
+	writeb(0x1, UDC_DEVCTL);
+
+	if (!(readb(UDC_DEVCTL) & 0x80)) {
+		printf("Error, the USB hardware is not on B mode\n");
+		writeb(0x0, UDC_DEVCTL);
+		return;
+	}
 }
 
-void udc_unset_nak(int ep_num)
+/* Turn off the USB connection */
+void udc_disconnect(void)
 {
-	/* noop */
+	if (!(readb(UDC_DEVCTL) & 0x80)) {
+		printf("Error, the USB hardware is not on B mode");
+		return;
+	}
+
+	writeb(0x0, UDC_DEVCTL);
 }
 
 int udc_endpoint_write(struct usb_endpoint_instance *endpoint)
 {
-	int ret = 0;
-
 	/* Transmit only if the hardware is available */
-	if (endpoint->tx_urb && endpoint->state == 0) {
-		unsigned int ep = endpoint->endpoint_address &
-			USB_ENDPOINT_NUMBER_MASK;
-
-		u16 peri_txcsr = readw(&musbr->ep[ep].epN.txcsr);
-
-		/* Error conditions */
-		if (peri_txcsr & MUSB_TXCSR_P_UNDERRUN) {
-			peri_txcsr &= ~MUSB_TXCSR_P_UNDERRUN;
-			writew(peri_txcsr, &musbr->ep[ep].epN.txcsr);
-		}
-
-		if (debug_level > 1)
-			musb_print_txcsr(peri_txcsr);
-
-		/* Check if a packet is waiting to be sent */
-		if (!(peri_txcsr & MUSB_TXCSR_TXPKTRDY)) {
-			u32 length;
-			u8 *data;
-			struct urb *urb = endpoint->tx_urb;
-			unsigned int remaining_packet = urb->actual_length -
-				endpoint->sent;
+	if (endpoint->tx_urb && endpoint->state == 0)
+		musb_ep_tx(endpoint);
 
-			if (endpoint->tx_packetSize < remaining_packet)
-				length = endpoint->tx_packetSize;
-			else
-				length = remaining_packet;
-
-			data = (u8 *) urb->buffer;
-			data += endpoint->sent;
+	return 0;
+}
 
-			/* common musb fifo function */
-			write_fifo(ep, length, data);
+/*
+ * udc_setup_ep - setup endpoint
+ *
+ * Associate a physical endpoint with endpoint_instance
+ */
+void udc_setup_ep(struct usb_device_instance *device, unsigned int ep,
+			struct usb_endpoint_instance *endpoint)
+{
+	int ep_addr;
+	int attributes;
 
-			musb_peri_tx_ready(ep);
+	/*
+	 * We dont' have a way to identify if the endpoint definitions changed,
+	 * so we have to always reconfigure the FIFOs to avoid problems
+	 */
+	musb_fifos_configure(device);
 
-			endpoint->last = length;
-			/* usbd_tx_complete will take care of updating 'sent' */
-			usbd_tx_complete(endpoint);
-		}
+	ep_addr = endpoint->endpoint_address;
+	if ((ep_addr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
+		/* IN endpoint */
+		attributes = endpoint->tx_attributes;
 	} else {
-		if (debug_level > 0)
-			serial_printf("ERROR : %s Problem with urb %p "
-				      "or ep state %d\n",
-				      __PRETTY_FUNCTION__,
-				      endpoint->tx_urb, endpoint->state);
+		/* OUT endpoint */
+		attributes = endpoint->rcv_attributes;
 	}
 
-	return ret;
-}
-
-void udc_setup_ep(struct usb_device_instance *device, unsigned int id,
-		  struct usb_endpoint_instance *endpoint)
-{
-	if (0 == id) {
-		/* EP0 */
-		ep0_endpoint = endpoint;
-		ep0_endpoint->endpoint_address = 0xff;
-		ep0_urb = usbd_alloc_urb(device, endpoint);
-	} else if (MAX_ENDPOINT >= id) {
-		int ep_addr;
-
-		/* Check the direction */
-		ep_addr = endpoint->endpoint_address;
-		if (USB_DIR_IN == (ep_addr & USB_ENDPOINT_DIR_MASK)) {
-			/* IN */
-			epinfo[(id * 2) + 1].epsize = endpoint->tx_packetSize;
-		} else {
-			/* OUT */
-			epinfo[id * 2].epsize = endpoint->rcv_packetSize;
+	writeb(ep & 0xF, UDC_INDEX);
+	if ((ep_addr & USB_ENDPOINT_DIR_MASK) == USB_DIR_IN) {
+		/* IN endpoint */
+		if (!ep) {		/* This only apply for ep != 0 */
+			/* Empty fifo twice on case of previous double buffer */
+			writew(1<<3, UDC_TXCSR);
+			writew(1<<3, UDC_TXCSR);
+
+			if (attributes & USB_ENDPOINT_XFER_ISOC)
+				writew(readw(UDC_TXCSR) | (1 << 13) | (1 << 14) |
+					0x4, UDC_TXCSR);
+			else
+				writew((readw(UDC_TXCSR) | (1 << 13) | 0x4) &
+					~(1 << 14), UDC_TXCSR);
 		}
-
-		musb_configure_ep(&epinfo[0],
-				  sizeof(epinfo) / sizeof(struct musb_epinfo));
+		/* Enable interrupt */
+		writew(readw(UDC_INTRTXE) | (1 << ep), UDC_INTRTXE);
 	} else {
-		if (debug_level > 0)
-			serial_printf("ERROR : %s endpoint request %d "
-				      "exceeds maximum %d\n",
-				      __PRETTY_FUNCTION__, id, MAX_ENDPOINT);
+		/* OUT endpoint */
+		if (!ep) {		/* This only apply for ep != 0 */
+			if (attributes & USB_ENDPOINT_XFER_ISOC)
+				writew((readw(UDC_RXCSR) | (1 << 14)) & ~(1 << 13),
+					UDC_RXCSR);
+			else
+				writew(readw(UDC_RXCSR) & ~(1 << 14) & ~(1 << 13),
+					UDC_RXCSR);
+		}
+		/* Enable interrupt */
+		writew(readw(UDC_INTRRXE) | (1 << ep), UDC_INTRRXE);
 	}
 }
 
-void udc_connect(void)
-{
-	/* noop */
-}
-
-void udc_disconnect(void)
-{
-	/* noop */
-}
-
-void udc_enable(struct usb_device_instance *device)
-{
-	/* Save the device structure pointer */
-	udc_device = device;
-
-	enabled = 1;
-}
-
-void udc_disable(void)
-{
-	enabled = 0;
-}
-
+/*
+ * udc_startup_events - allow udc code to do any additional startup
+ */
 void udc_startup_events(struct usb_device_instance *device)
 {
 	/* The DEVICE_INIT event puts the USB device in the state STATE_INIT. */
@@ -931,35 +698,86 @@ void udc_startup_events(struct usb_devic
 	 */
 	usbd_device_event_irq(device, DEVICE_CREATE, 0);
 
-	/* Resets the address to 0 */
-	usbd_device_event_irq(device, DEVICE_RESET, 0);
+	/*
+	 * Some USB controller driver implementations signal
+	 * DEVICE_HUB_CONFIGURED and DEVICE_RESET events here.
+	 * DEVICE_HUB_CONFIGURED causes a transition to the state STATE_POWERED,
+	 * and DEVICE_RESET causes a transition to the state STATE_DEFAULT.
+	 * The MUSB client controller has the capability to detect when the
+	 * USB cable is connected to a powered USB bus, so we will defer the
+	 * DEVICE_HUB_CONFIGURED and DEVICE_RESET events until later.
+	 */
+
+	/* Save the device structure pointer */
+	udc_device = device;
 
-	udc_enable(device);
+	/* Setup ep0 urb */
+	if (!ep0_urb) {
+		ep0_urb =
+			usbd_alloc_urb(udc_device, udc_device->bus->
+					endpoint_array);
+	} else {
+		printf("udc_enable: ep0_urb already allocated %p\n", ep0_urb);
+	}
+
+	/* Enable control interrupts */
+	writeb(0xf7, UDC_INTRUSBE);
 }
 
-int udc_init(void)
+void udc_set_nak(int epid)
 {
-	int ret;
-	int ep_loop;
+/*
+ * On MUSB the NAKing is controlled by the USB controller buffers,
+ * so as long as we don't read data from the FIFO, the controller will NAK.
+ * Nothing to see here, move along...
+ */
+}
 
-	ret = musb_platform_init();
-	if (ret < 0)
-		goto end;
+void udc_unset_nak(int epid)
+{
+/*
+ * On MUSB the NAKing is controlled by the USB controller buffers,
+ * so as long as we don't read data from the FIFO, the controller will NAK.
+ * Nothing to see here, move along...
+ */
+}
 
-	/* Configure all the endpoint FIFO's and start usb controller */
-	musbr = musb_cfg.regs;
+/* Start to initialize h/w stuff */
+int udc_init(void)
+{
+	/* Clock is initialized on the board code */
 
-	/* Initialize the endpoints */
-	for (ep_loop = 0; ep_loop < MAX_ENDPOINT * 2; ep_loop++) {
-		epinfo[ep_loop].epnum = (ep_loop / 2) + 1;
-		epinfo[ep_loop].epdir = ep_loop % 2; /* OUT, IN */
-		epinfo[ep_loop].epsize = 0;
-	}
+	if (musb_platform_init() == -1)
+		return -1;
 
-	musb_peri_softconnect();
+	/* MUSB soft-reset */
+	writel(2, UDC_SYSCONFIG);
+	/* MUSB end any previous session */
+	writeb(0x0, UDC_DEVCTL);
+
+	writel(readl(UDC_FORCESTDBY) & ~1, UDC_FORCESTDBY); /* disable MSTANDBY */
+	writel(readl(UDC_SYSCONFIG) | (2<<12), UDC_SYSCONFIG); /* ena SMARTSTDBY */
+	writel(readl(UDC_SYSCONFIG) & ~1, UDC_SYSCONFIG); /* disable AUTOIDLE */
+	writel(readl(UDC_SYSCONFIG) | (2<<3), UDC_SYSCONFIG); /* enable SMARTIDLE */
+	writel(readl(UDC_SYSCONFIG) | 1, UDC_SYSCONFIG); /* enable AUTOIDLE */
+
+	/* Configure the PHY as PHY interface is 12-pin, 8-bit SDR ULPI */
+	sr32((void *)UDC_INTERFSEL, 0, 1, 1);
+
+	/* Turn off interrupts */
+	writew(0x00, UDC_INTRTXE);
+	writew(0x00, UDC_INTRRXE);
 
-	ret = 0;
-end:
+#if CONFIG_MUSB_FULL_SPEED
+	/*
+	 * Use Full speed for debugging proposes, useful so most USB
+	 * analyzers can catch the transactions
+	 */
+	writeb(0, UDC_POWER);
+	printf("MUSB: using full speed\n");
+#else
+	printf("MUSB: using high speed\n");
+#endif
 
-	return ret;
+	return 0;
 }
diff -ruNpa u-boot-2010.12/include/configs/omap3_beagle.h u-boot-2010.12.new/include/configs/omap3_beagle.h
--- u-boot-2010.12/include/configs/omap3_beagle.h	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/include/configs/omap3_beagle.h	2011-03-01 11:46:54.491549998 +1100
@@ -121,6 +121,10 @@
 #define CONFIG_USB_TTY			1
 #define CONFIG_SYS_CONSOLE_IS_IN_ENV	1
 
+#define CONFIG_USBD_DFU			1
+#define CONFIG_USBD_DFU_XFER_SIZE 	0x1000
+#define CONFIG_USBD_DFU_INTERFACE	2
+
 /* commands to include */
 #include <config_cmd_default.h>
 
@@ -227,6 +231,7 @@
 		"nand read ${loadaddr} 280000 400000; " \
 		"bootm ${loadaddr}\0" \
 
+#if 0
 #define CONFIG_BOOTCOMMAND \
 	"if mmc rescan ${mmcdev}; then " \
 		"if run loadbootscript; then " \
@@ -238,6 +243,7 @@
 			"fi; " \
 		"fi; " \
 	"else run nandboot; fi"
+#endif
 
 #define CONFIG_AUTO_COMPLETE		1
 /*
@@ -279,7 +285,7 @@
  */
 #define CONFIG_STACKSIZE	(128 << 10)	/* regular stack 128 KiB */
 #ifdef CONFIG_USE_IRQ
-#define CONFIG_STACKSIZE_IRQ	(4 << 10)	/* IRQ stack 4 KiB */
+#define CONFIG_STACKSIZE_IRQ	(8 << 10)	/* IRQ stack 4 KiB */
 #define CONFIG_STACKSIZE_FIQ	(4 << 10)	/* FIQ stack 4 KiB */
 #endif
 
diff -ruNpa u-boot-2010.12/include/usb/musb_udc.h u-boot-2010.12.new/include/usb/musb_udc.h
--- u-boot-2010.12/include/usb/musb_udc.h	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/include/usb/musb_udc.h	2011-03-01 11:48:06.811549998 +1100
@@ -1,11 +1,15 @@
 /*
- * Copyright (c) 2009 Wind River Systems, Inc.
- * Tom Rix <Tom.Rix@windriver.com>
+ * (C) Copyright 2009 Texas Instruments Incorporated.
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 2 of
- * the License, or (at your option) any later version.
+ * Based on
+ * u-boot OMAP1510 USB drivers (include/usbdcore_omap1510.h)
+ *
+ * Author: Diego Dompe (diego.dompe@ridgerun.com)
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2 of the License, or
+ * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -14,40 +18,101 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
- * MA 02111-1307 USA
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  */
-#ifndef __MUSB_UDC_H__
-#define __MUSB_UDC_H__
+
+#ifndef __USBDCORE_MUSB_H__
+#define __USBDCORE_MUSB_H__
 
 #include <usbdevice.h>
 
-/* UDC level routines */
+/* USB Function Module Registers */
+
+/*
+ * UDC_BASE is defined for the specific silicon under soc
+ * specific cpu.h or related header.
+ */
+
+#define UDC_OFFSET(offset)	(UDC_BASE+(offset))
+
+#define UDC_INTSRCR		UDC_OFFSET(0x0A) /* USB Interrupt src reg */
+#define UDC_INTCLRR		UDC_OFFSET(0x0A) /* USB Interrupt src clr reg */
+
+#define UDC_FADDR		UDC_OFFSET(0x00)
+#define UDC_POWER		UDC_OFFSET(0x01)
+#define UDC_INTRTX		UDC_OFFSET(0x02)
+#define UDC_INTRRX		UDC_OFFSET(0x04)
+#define UDC_INTRTXE		UDC_OFFSET(0x06) /* Enable reg for INTRTX */
+#define UDC_INTRRXE		UDC_OFFSET(0x08) /* Enable reg for INTRRX */
+#define UDC_INTRUSB		UDC_OFFSET(0x0A)
+#define UDC_INTRUSBE		UDC_OFFSET(0x0B)
+#define UDC_INDEX		UDC_OFFSET(0x0E)
+#define UDC_TESTMODE		UDC_OFFSET(0x0F)
+#define UDC_TXMAXP		UDC_OFFSET(0x10)
+#define UDC_CSR0		UDC_OFFSET(0x12)
+#define UDC_TXCSR		UDC_OFFSET(0x12)
+#define UDC_RXMAXP		UDC_OFFSET(0x14)
+#define UDC_RXCSR		UDC_OFFSET(0x16)
+#define UDC_COUNT0		UDC_OFFSET(0x18)
+#define UDC_RXCOUNT		UDC_OFFSET(0x18)
+#define UDC_FIFO0		UDC_OFFSET(0x20)
+#define UDC_FIFO1		UDC_OFFSET(0x24)
+#define UDC_FIFO2		UDC_OFFSET(0x28)
+#define UDC_FIFO3		UDC_OFFSET(0x2C)
+#define UDC_FIFO4		UDC_OFFSET(0x30)
+#define UDC_FIFO5		UDC_OFFSET(0x34)
+#define UDC_FIFO6		UDC_OFFSET(0x38)
+#define UDC_FIFO7		UDC_OFFSET(0x3C)
+#define UDC_FIFO8		UDC_OFFSET(0x40)
+#define UDC_FIFO9		UDC_OFFSET(0x44)
+#define UDC_FIFO10		UDC_OFFSET(0x48)
+#define UDC_FIFO11		UDC_OFFSET(0x4C)
+#define UDC_FIFO12		UDC_OFFSET(0x50)
+#define UDC_FIFO13		UDC_OFFSET(0x54)
+#define UDC_FIFO14		UDC_OFFSET(0x58)
+#define UDC_FIFO15		UDC_OFFSET(0x5C)
+#define UDC_DEVCTL		UDC_OFFSET(0x60)
+#define UDC_TXFIFOSZ		UDC_OFFSET(0x62)
+#define UDC_RXFIFOSZ		UDC_OFFSET(0x63)
+#define UDC_TXFIFOADDR		UDC_OFFSET(0x64)
+#define UDC_RXFIFOADDR		UDC_OFFSET(0x66)
+
+#define UDC_SYSCONFIG		UDC_OFFSET(0x404)
+#define UDC_INTERFSEL		UDC_OFFSET(0x40C)
+#define UDC_FORCESTDBY		UDC_OFFSET(0x414)
+
+/* MUSB Endpoint parameters */
+#define EP0_MAX_PACKET_SIZE	64
+#define UDC_OUT_ENDPOINT	3	/* Device RX endpoint */
+#define UDC_OUT_PACKET_SIZE	512
+#define UDC_IN_ENDPOINT		2	/* Device TX endpoint */
+#define UDC_IN_PACKET_SIZE	512
+#define UDC_INT_ENDPOINT	1	/* Device Interrupt/Status endpoint */
+#define UDC_INT_PACKET_SIZE	16
+#define UDC_BULK_PACKET_SIZE	512
+
+#define UDC_MAX_FIFO_SIZE	16384
+
+#define DEV_CONFIG_VALUE	1	/* Only one i.e. CDC */
+
 void udc_irq(void);
-void udc_set_nak(int ep_num);
-void udc_unset_nak(int ep_num);
+/* Flow control */
+void udc_set_nak(int epid);
+void udc_unset_nak(int epid);
+
+/* Higher level functions for abstracting away from specific device */
 int udc_endpoint_write(struct usb_endpoint_instance *endpoint);
-void udc_setup_ep(struct usb_device_instance *device, unsigned int id,
-		  struct usb_endpoint_instance *endpoint);
-void udc_connect(void);
-void udc_disconnect(void);
-void udc_enable(struct usb_device_instance *device);
-void udc_disable(void);
-void udc_startup_events(struct usb_device_instance *device);
+
 int udc_init(void);
 
-/* usbtty */
-#ifdef CONFIG_USB_TTY
+void udc_enable(struct usb_device_instance *device);
+void udc_disable(void);
 
-#define EP0_MAX_PACKET_SIZE	64 /* MUSB_EP0_FIFOSIZE */
-#define UDC_INT_ENDPOINT	1
-#define UDC_INT_PACKET_SIZE	64
-#define UDC_OUT_ENDPOINT	2
-#define UDC_OUT_PACKET_SIZE	64
-#define UDC_IN_ENDPOINT		3
-#define UDC_IN_PACKET_SIZE	64
-#define UDC_BULK_PACKET_SIZE	64
+void udc_connect(void);
+void udc_disconnect(void);
 
-#endif /* CONFIG_USB_TTY */
+void udc_startup_events(struct usb_device_instance *device);
+void udc_setup_ep(struct usb_device_instance *device, unsigned int ep,
+			struct usb_endpoint_instance *endpoint);
 
-#endif /* __MUSB_UDC_H__ */
+#endif /* __USBDCORE_MUSB_H__ */
diff -ruNpa u-boot-2010.12/include/usbdevice.h u-boot-2010.12.new/include/usbdevice.h
--- u-boot-2010.12/include/usbdevice.h	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/include/usbdevice.h	2011-03-01 11:48:06.811549998 +1100
@@ -33,6 +33,7 @@
 
 #include <common.h>
 #include "usbdescriptors.h"
+#include <usb_dfu_descriptors.h>
 
 
 #define MAX_URBS_QUEUED 5
@@ -467,7 +468,11 @@ typedef struct urb_link {
  * function driver to inform it that data has arrived.
  */
 
+#ifdef CONFIG_USBD_DFU
+#define URB_BUF_SIZE (128+CONFIG_USBD_DFU_XFER_SIZE)
+#else
 #define URB_BUF_SIZE 128 /* in linux we'd malloc this, but in u-boot we prefer static data */
+#endif
 struct urb {
 
 	struct usb_endpoint_instance *endpoint;
@@ -595,6 +600,12 @@ struct usb_device_instance {
 	unsigned long usbd_rxtx_timestamp;
 	unsigned long usbd_last_rxtx_timestamp;
 
+#ifdef CONFIG_USBD_DFU
+	const struct usb_device_descriptor *dfu_dev_desc;
+	const struct _dfu_desc *dfu_cfg_desc;
+	enum dfu_state dfu_state;
+	u_int8_t dfu_status;
+#endif
 };
 
 /* Bus Interface configuration structure
@@ -624,6 +635,8 @@ extern char *usbd_device_status[];
 extern char *usbd_device_requests[];
 extern char *usbd_device_descriptors[];
 
+extern struct usb_string_descriptor **usb_strings;
+
 void urb_link_init (urb_link * ul);
 void urb_detach (struct urb *urb);
 urb_link *first_urb_link (urb_link * hd);
diff -ruNpa u-boot-2010.12/include/usb_dfu_descriptors.h u-boot-2010.12.new/include/usb_dfu_descriptors.h
--- u-boot-2010.12/include/usb_dfu_descriptors.h	1970-01-01 10:00:00.000000000 +1000
+++ u-boot-2010.12.new/include/usb_dfu_descriptors.h	2011-03-01 11:48:06.811549998 +1100
@@ -0,0 +1,94 @@
+#ifndef _USB_DFU_H
+#define _USB_DFU_H
+/* USB Device Firmware Update Implementation for OpenPCD
+ * (C) 2006 by Harald Welte <hwelte@hmw-consulting.de>
+ *
+ * Protocol definitions for USB DFU
+ *
+ * This ought to be compliant to the USB DFU Spec 1.0 as available from
+ * http://www.usb.org/developers/devclass_docs/usbdfu10.pdf
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
+
+#include <linux/types.h>
+
+#define USB_DT_DFU			0x21
+
+struct usb_dfu_func_descriptor {
+	u_int8_t		bLength;
+	u_int8_t		bDescriptorType;
+	u_int8_t		bmAttributes;
+#define USB_DFU_CAN_DOWNLOAD	(1 << 0)
+#define USB_DFU_CAN_UPLOAD	(1 << 1)
+#define USB_DFU_MANIFEST_TOL	(1 << 2)
+#define USB_DFU_WILL_DETACH	(1 << 3)
+	u_int16_t		wDetachTimeOut;
+	u_int16_t		wTransferSize;
+	u_int16_t		bcdDFUVersion;
+} __attribute__ ((packed));
+
+#define USB_DT_DFU_SIZE			9
+
+#define USB_TYPE_DFU		(USB_TYPE_CLASS|USB_RECIP_INTERFACE)
+
+/* DFU class-specific requests (Section 3, DFU Rev 1.1) */
+#define USB_REQ_DFU_DETACH	0x00
+#define USB_REQ_DFU_DNLOAD	0x01
+#define USB_REQ_DFU_UPLOAD	0x02
+#define USB_REQ_DFU_GETSTATUS	0x03
+#define USB_REQ_DFU_CLRSTATUS	0x04
+#define USB_REQ_DFU_GETSTATE	0x05
+#define USB_REQ_DFU_ABORT	0x06
+
+struct dfu_status {
+	u_int8_t bStatus;
+	u_int8_t bwPollTimeout[3];
+	u_int8_t bState;
+	u_int8_t iString;
+} __attribute__((packed));
+
+#define DFU_STATUS_OK			0x00
+#define DFU_STATUS_errTARGET		0x01
+#define DFU_STATUS_errFILE		0x02
+#define DFU_STATUS_errWRITE		0x03
+#define DFU_STATUS_errERASE		0x04
+#define DFU_STATUS_errCHECK_ERASED	0x05
+#define DFU_STATUS_errPROG		0x06
+#define DFU_STATUS_errVERIFY		0x07
+#define DFU_STATUS_errADDRESS		0x08
+#define DFU_STATUS_errNOTDONE		0x09
+#define DFU_STATUS_errFIRMWARE		0x0a
+#define DFU_STATUS_errVENDOR		0x0b
+#define DFU_STATUS_errUSBR		0x0c
+#define DFU_STATUS_errPOR		0x0d
+#define DFU_STATUS_errUNKNOWN		0x0e
+#define DFU_STATUS_errSTALLEDPKT	0x0f
+
+enum dfu_state {
+	DFU_STATE_appIDLE		= 0,
+	DFU_STATE_appDETACH		= 1,
+	DFU_STATE_dfuIDLE		= 2,
+	DFU_STATE_dfuDNLOAD_SYNC	= 3,
+	DFU_STATE_dfuDNBUSY		= 4,
+	DFU_STATE_dfuDNLOAD_IDLE	= 5,
+	DFU_STATE_dfuMANIFEST_SYNC	= 6,
+	DFU_STATE_dfuMANIFEST		= 7,
+	DFU_STATE_dfuMANIFEST_WAIT_RST	= 8,
+	DFU_STATE_dfuUPLOAD_IDLE	= 9,
+	DFU_STATE_dfuERROR		= 10,
+};
+
+#endif /* _USB_DFU_H */
diff -ruNpa u-boot-2010.12/include/usb_dfu.h u-boot-2010.12.new/include/usb_dfu.h
--- u-boot-2010.12/include/usb_dfu.h	1970-01-01 10:00:00.000000000 +1000
+++ u-boot-2010.12.new/include/usb_dfu.h	2011-03-01 11:48:06.811549998 +1100
@@ -0,0 +1,95 @@
+#ifndef _DFU_H
+#define _DFU_H
+
+/* USB Device Firmware Update Implementation for u-boot
+ * (C) 2007 by OpenMoko, Inc.
+ * Author: Harald Welte <laforge@openmoko.org>
+ *
+ * based on: USB Device Firmware Update Implementation for OpenPCD
+ * (C) 2006 by Harald Welte <hwelte@hmw-consulting.de>
+ *
+ * This ought to be compliant to the USB DFU Spec 1.0 as available from
+ * http://www.usb.org/developers/devclass_docs/usbdfu10.pdf
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
+
+#include <asm/types.h>
+#include <usbdevice.h>
+#include <usbdescriptors.h>
+#include <usb_dfu_descriptors.h>
+#include <config.h>
+
+/* USB DFU functional descriptor */
+#define DFU_FUNC_DESC  {						\
+	.bLength		= USB_DT_DFU_SIZE,			\
+	.bDescriptorType	= USB_DT_DFU,				\
+	.bmAttributes		= USB_DFU_CAN_UPLOAD | USB_DFU_CAN_DOWNLOAD | USB_DFU_MANIFEST_TOL, \
+	.wDetachTimeOut		= 0xff00,				\
+	.wTransferSize		= CONFIG_USBD_DFU_XFER_SIZE,		\
+	.bcdDFUVersion		= 0x0100,				\
+}
+
+/* USB Interface descriptor in Runtime mode */
+#define DFU_RT_IF_DESC	{						\
+	.bLength		= USB_DT_INTERFACE_SIZE,		\
+	.bDescriptorType	= USB_DT_INTERFACE,			\
+	.bInterfaceNumber	= CONFIG_USBD_DFU_INTERFACE,		\
+	.bAlternateSetting	= 0x00,					\
+	.bNumEndpoints		= 0x00,					\
+	.bInterfaceClass	= 0xfe,					\
+	.bInterfaceSubClass	= 0x01,					\
+	.bInterfaceProtocol	= 0x01,					\
+	.iInterface		= DFU_STR_CONFIG,			\
+}
+
+#define ARRAY_SIZE(x)           (sizeof(x) / sizeof((x)[0]))
+
+#ifndef DFU_NUM_ALTERNATES
+#define DFU_NUM_ALTERNATES	6
+#endif
+
+#define DFU_STR_MANUFACTURER	STR_MANUFACTURER
+#define DFU_STR_PRODUCT		STR_PRODUCT
+#define DFU_STR_SERIAL		STR_SERIAL
+#define DFU_STR_CONFIG		(STR_COUNT)
+#define DFU_STR_ALT(n)		(STR_COUNT+(n)+1)
+#define DFU_STR_COUNT		DFU_STR_ALT(DFU_NUM_ALTERNATES)
+
+#define CONFIG_DFU_CFG_STR	"USB Device Firmware Upgrade"
+#define CONFIG_DFU_ALT0_STR	"RAM 0x81000000"
+
+struct _dfu_desc {
+	struct usb_configuration_descriptor ucfg;
+	struct usb_interface_descriptor uif[DFU_NUM_ALTERNATES];
+	struct usb_dfu_func_descriptor func_dfu;
+};
+
+int dfu_init_instance(struct usb_device_instance *dev);
+
+#define DFU_EP0_NONE		0
+#define DFU_EP0_UNHANDLED	1
+#define DFU_EP0_STALL		2
+#define DFU_EP0_ZLP		3
+#define DFU_EP0_DATA		4
+
+extern volatile enum dfu_state *system_dfu_state; /* for 3rd parties */
+
+int dfu_ep0_handler(struct urb *urb);
+
+void dfu_event(struct usb_device_instance *device,
+	       usb_device_event_t event, int data);
+
+#endif /* _DFU_H */
diff -ruNpa u-boot-2010.12/include/usb_dfu_trailer.h u-boot-2010.12.new/include/usb_dfu_trailer.h
--- u-boot-2010.12/include/usb_dfu_trailer.h	1970-01-01 10:00:00.000000000 +1000
+++ u-boot-2010.12.new/include/usb_dfu_trailer.h	2011-03-01 11:48:06.811549998 +1100
@@ -0,0 +1,31 @@
+#ifndef _USB_DFU_TRAILER_H
+#define _USB_DFU_TRAILER_H
+
+/* trailer handling for DFU files */
+
+#define UBOOT_DFU_TRAILER_V1	1
+#define UBOOT_DFU_TRAILER_MAGIC	0x19731978
+struct uboot_dfu_trailer {
+	u_int32_t	magic;
+	u_int16_t	version;
+	u_int16_t	length;
+	u_int16_t	vendor;
+	u_int16_t	product;
+	u_int32_t	revision;
+} __attribute__((packed));
+
+/* we mirror the trailer because we want it to be longer in later versions
+ * while keeping backwards compatibility */
+static inline void dfu_trailer_mirror(struct uboot_dfu_trailer *trailer,
+				      unsigned char *eof)
+{
+	int i;
+	int len = sizeof(struct uboot_dfu_trailer);
+	unsigned char *src = eof - len;
+	unsigned char *dst = (unsigned char *) trailer;
+
+	for (i = 0; i < len; i++)
+		dst[len-1-i] = src[i];
+}
+
+#endif /* _USB_DFU_TRAILER_H */
diff -ruNpa u-boot-2010.12/Makefile u-boot-2010.12.new/Makefile
--- u-boot-2010.12/Makefile	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/Makefile	2011-03-01 11:48:06.811549998 +1100
@@ -345,6 +345,12 @@ $(obj)u-boot.ldr.hex:	$(obj)u-boot.ldr
 $(obj)u-boot.ldr.srec:	$(obj)u-boot.ldr
 		$(OBJCOPY) ${OBJCFLAGS} -O srec $< $@ -I binary
 
+$(obj)u-boot.udfu:	$(obj)u-boot.bin
+		./tools/mkudfu -v $(CONFIG_USB_DFU_VENDOR) \
+			       -p $(CONFIG_USB_DFU_PRODUCT) \
+			       -r $(CONFIG_USB_DFU_REVISION) \
+			       -d $< $@
+
 $(obj)u-boot.img:	$(obj)u-boot.bin
 		$(obj)tools/mkimage -A $(ARCH) -T firmware -C none \
 		-a $(CONFIG_SYS_TEXT_BASE) -e 0 \
diff -ruNpa u-boot-2010.12/tools/Makefile u-boot-2010.12.new/tools/Makefile
--- u-boot-2010.12/tools/Makefile	2010-12-23 06:22:14.000000000 +1100
+++ u-boot-2010.12.new/tools/Makefile	2011-03-01 11:48:06.811549998 +1100
@@ -65,6 +65,7 @@ BIN_FILES-$(CONFIG_INCA_IP) += inca-swap
 BIN_FILES-y += mkimage$(SFX)
 BIN_FILES-$(CONFIG_NETCONSOLE) += ncb$(SFX)
 BIN_FILES-$(CONFIG_SHA1_CHECK_UB_IMG) += ubsha1$(SFX)
+BIN_FILES-$(CONFIG_USBD_DFU) += mkudfu$(SFX)
 
 # Source files which exist outside the tools directory
 EXT_OBJ_FILES-y += common/env_embedded.o
@@ -88,6 +89,7 @@ NOPED_OBJ_FILES-y += mkimage.o
 OBJ_FILES-$(CONFIG_NETCONSOLE) += ncb.o
 NOPED_OBJ_FILES-y += os_support.o
 OBJ_FILES-$(CONFIG_SHA1_CHECK_UB_IMG) += ubsha1.o
+OBJ_FILES-$(CONFIG_USBD_DFU) += mkudfu.o
 
 # Don't build by default
 #ifeq ($(ARCH),ppc)
@@ -193,6 +195,10 @@ $(obj)mpc86x_clk$(SFX):	$(obj)mpc86x_clk
 	$(HOSTCC) $(HOSTCFLAGS) $(HOSTLDFLAGS) -o $@ $^
 	$(HOSTSTRIP) $@
 
+$(obj)mkudfu$(SFX):	$(obj)mkudfu.o
+	$(HOSTCC) $(HOSTCFLAGS) $(HOST_LDFLAGS) -o $@ $^
+	$(HOSTSTRIP) $@
+
 $(obj)ncb$(SFX):	$(obj)ncb.o
 	$(HOSTCC) $(HOSTCFLAGS) $(HOSTLDFLAGS) -o $@ $^
 	$(HOSTSTRIP) $@
diff -ruNpa u-boot-2010.12/tools/mkudfu.c u-boot-2010.12.new/tools/mkudfu.c
--- u-boot-2010.12/tools/mkudfu.c	1970-01-01 10:00:00.000000000 +1000
+++ u-boot-2010.12.new/tools/mkudfu.c	2011-03-01 11:48:06.811549998 +1100
@@ -0,0 +1,314 @@
+/*
+ * USB DFU file trailer tool
+ * (C) Copyright by OpenMoko, Inc.
+ * Author: Harald Welte <laforge@openmoko.org>
+ *
+ * based on mkimage.c, copyright information as follows:
+ *
+ * (C) Copyright 2000-2004
+ * DENX Software Engineering
+ * Wolfgang Denk, wd@denx.de
+ * All rights reserved.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of
+ * the License, or (at your option) any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
+ * MA 02111-1307 USA
+ */
+
+#include <errno.h>
+#include <fcntl.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#ifndef __WIN32__
+#include <netinet/in.h>		/* for host / network byte order conversions	*/
+#endif
+#include <sys/mman.h>
+#include <sys/stat.h>
+#include <time.h>
+#include <unistd.h>
+
+#if defined(__BEOS__) || defined(__NetBSD__) || defined(__APPLE__)
+#include <inttypes.h>
+#endif
+
+#ifdef __WIN32__
+typedef unsigned int __u32;
+
+#define SWAP_LONG(x) \
+	((__u32)( \
+		(((__u32)(x) & (__u32)0x000000ffUL) << 24) | \
+		(((__u32)(x) & (__u32)0x0000ff00UL) <<  8) | \
+		(((__u32)(x) & (__u32)0x00ff0000UL) >>  8) | \
+		(((__u32)(x) & (__u32)0xff000000UL) >> 24) ))
+typedef		unsigned char	uint8_t;
+typedef		unsigned short	uint16_t;
+typedef		unsigned int	uint32_t;
+
+#define     ntohl(a)	SWAP_LONG(a)
+#define     htonl(a)	SWAP_LONG(a)
+#endif	/* __WIN32__ */
+
+#ifndef	O_BINARY		/* should be define'd on __WIN32__ */
+#define O_BINARY	0
+#endif
+
+#include <usb_dfu_trailer.h>
+
+extern int errno;
+
+#ifndef MAP_FAILED
+#define MAP_FAILED (-1)
+#endif
+
+static char *cmdname;
+
+static char *datafile;
+static char *imagefile;
+
+
+static void usage()
+{
+	fprintf (stderr, "%s - create / display u-boot DFU trailer\n", cmdname);
+	fprintf (stderr, "Usage: %s -l image\n"
+			 "          -l ==> list image header information\n"
+			 "       %s -v VID -p PID -r REV -d data_file image\n",
+		cmdname, cmdname);
+	fprintf (stderr, "          -v ==> set vendor ID to 'VID'\n"
+			 "          -p ==> set product ID system to 'PID'\n"
+			 "          -r ==> set hardware revision to 'REV'\n"
+			 "          -d ==> use 'data_file' as input file\n"
+		);
+	exit (EXIT_FAILURE);
+}
+
+static void print_trailer(struct uboot_dfu_trailer *trailer)
+{
+	printf("===> DFU Trailer information:\n");
+	printf("Trailer Vers.:	%d\n", trailer->version);
+	printf("Trailer Length:	%d\n", trailer->length);
+	printf("VendorID:	0x%04x\n", trailer->vendor);
+	printf("ProductID:	0x%04x\n", trailer->product);
+	printf("HW Revision:	0x%04x\n", trailer->revision);
+}
+
+static void copy_file (int ifd, const char *datafile, int pad)
+{
+	int dfd;
+	struct stat sbuf;
+	unsigned char *ptr;
+	int tail;
+	int zero = 0;
+	int offset = 0;
+	int size;
+
+	if ((dfd = open(datafile, O_RDONLY|O_BINARY)) < 0) {
+		fprintf (stderr, "%s: Can't open %s: %s\n",
+			cmdname, datafile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	if (fstat(dfd, &sbuf) < 0) {
+		fprintf (stderr, "%s: Can't stat %s: %s\n",
+			cmdname, datafile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	ptr = (unsigned char *)mmap(0, sbuf.st_size,
+				    PROT_READ, MAP_SHARED, dfd, 0);
+	if (ptr == (unsigned char *)MAP_FAILED) {
+		fprintf (stderr, "%s: Can't read %s: %s\n",
+			cmdname, datafile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	size = sbuf.st_size - offset;
+	if (write(ifd, ptr + offset, size) != size) {
+		fprintf (stderr, "%s: Write error on %s: %s\n",
+			cmdname, imagefile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	if (pad && ((tail = size % 4) != 0)) {
+
+		if (write(ifd, (char *)&zero, 4-tail) != 4-tail) {
+			fprintf (stderr, "%s: Write error on %s: %s\n",
+				cmdname, imagefile, strerror(errno));
+			exit (EXIT_FAILURE);
+		}
+	}
+
+	(void) munmap((void *)ptr, sbuf.st_size);
+	(void) close (dfd);
+}
+
+
+int main(int argc, char **argv)
+{
+	int ifd;
+	int lflag = 0;
+	struct stat sbuf;
+	u_int16_t opt_vendor, opt_product, opt_revision;
+	struct uboot_dfu_trailer _hdr, _mirror, *hdr = &_hdr;
+
+	opt_vendor = opt_product = opt_revision = 0;
+
+	cmdname = *argv;
+
+	while (--argc > 0 && **++argv == '-') {
+		while (*++*argv) {
+			switch (**argv) {
+			case 'l':
+				lflag = 1;
+				break;
+			case 'v':
+				if (--argc <= 0)
+					usage ();
+				opt_vendor = strtoul(*++argv, NULL, 16);
+				goto NXTARG;
+			case 'p':
+				if (--argc <= 0)
+					usage ();
+				opt_product = strtoul(*++argv, NULL, 16);
+				goto NXTARG;
+			case 'r':
+				if (--argc <= 0)
+					usage ();
+				opt_revision = strtoul(*++argv, NULL, 16);
+				goto NXTARG;
+			case 'd':
+				if (--argc <= 0)
+					usage ();
+				datafile = *++argv;
+				goto NXTARG;
+			case 'h':
+				usage();
+				break;
+			default:
+				usage();
+			}
+		}
+NXTARG:		;
+	}
+
+	if (argc != 1)
+		usage();
+
+	imagefile = *argv;
+
+	if (lflag)
+		ifd = open(imagefile, O_RDONLY|O_BINARY);
+	else
+		ifd = open(imagefile, O_RDWR|O_CREAT|O_TRUNC|O_BINARY, 0666);
+
+	if (ifd < 0) {
+		fprintf (stderr, "%s: Can't open %s: %s\n",
+			cmdname, imagefile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	if (lflag) {
+		unsigned char *ptr;
+		/* list header information of existing image */
+		if (fstat(ifd, &sbuf) < 0) {
+			fprintf (stderr, "%s: Can't stat %s: %s\n",
+				cmdname, imagefile, strerror(errno));
+			exit (EXIT_FAILURE);
+		}
+
+		if ((unsigned)sbuf.st_size < sizeof(struct uboot_dfu_trailer)) {
+			fprintf (stderr,
+				"%s: Bad size: \"%s\" is no valid image\n",
+				cmdname, imagefile);
+			exit (EXIT_FAILURE);
+		}
+
+		ptr = (unsigned char *)mmap(0, sbuf.st_size,
+					    PROT_READ, MAP_SHARED, ifd, 0);
+		if ((caddr_t)ptr == (caddr_t)-1) {
+			fprintf (stderr, "%s: Can't read %s: %s\n",
+				cmdname, imagefile, strerror(errno));
+			exit (EXIT_FAILURE);
+		}
+
+		dfu_trailer_mirror(hdr, ptr+sbuf.st_size);
+
+		if (hdr->magic != UBOOT_DFU_TRAILER_MAGIC) {
+			fprintf (stderr,
+				"%s: Bad Magic Number: \"%s\" is no valid image\n",
+				cmdname, imagefile);
+			exit (EXIT_FAILURE);
+		}
+
+		/* for multi-file images we need the data part, too */
+		print_trailer(hdr);
+
+		(void) munmap((void *)ptr, sbuf.st_size);
+		(void) close (ifd);
+
+		exit (EXIT_SUCCESS);
+	}
+
+	/* if we're not listing: */
+
+	copy_file (ifd, datafile, 0);
+
+	memset (hdr, 0, sizeof(struct uboot_dfu_trailer));
+
+	/* Build new header */
+	hdr->version	= UBOOT_DFU_TRAILER_V1;
+	hdr->magic	= UBOOT_DFU_TRAILER_MAGIC;
+	hdr->length	= sizeof(struct uboot_dfu_trailer);
+	hdr->vendor	= opt_vendor;
+	hdr->product	= opt_product;
+	hdr->revision	= opt_revision;
+
+	print_trailer(hdr);
+	dfu_trailer_mirror(&_mirror, (unsigned char *)hdr+sizeof(*hdr));
+
+	if (write(ifd, &_mirror, sizeof(struct uboot_dfu_trailer))
+					!= sizeof(struct uboot_dfu_trailer)) {
+		fprintf (stderr, "%s: Write error on %s: %s\n",
+			cmdname, imagefile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	/* We're a bit of paranoid */
+#if defined(_POSIX_SYNCHRONIZED_IO) && !defined(__sun__) && !defined(__FreeBSD__)
+	(void) fdatasync (ifd);
+#else
+	(void) fsync (ifd);
+#endif
+
+	if (fstat(ifd, &sbuf) < 0) {
+		fprintf (stderr, "%s: Can't stat %s: %s\n",
+			cmdname, imagefile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	/* We're a bit of paranoid */
+#if defined(_POSIX_SYNCHRONIZED_IO) && !defined(__sun__) && !defined(__FreeBSD__)
+	(void) fdatasync (ifd);
+#else
+	(void) fsync (ifd);
+#endif
+
+	if (close(ifd)) {
+		fprintf (stderr, "%s: Write error on %s: %s\n",
+			cmdname, imagefile, strerror(errno));
+		exit (EXIT_FAILURE);
+	}
+
+	exit (EXIT_SUCCESS);
+}
